"""
# Base ElabPass
"""


# Std-Lib Imports
import sys
from pathlib import Path
from dataclasses import field
from typing import Dict, List, Set, Optional, Union

# Local imports
from ...datatype import datatype, AllowArbConfig
from ...module import Module
from ...external_module import ExternalModuleCall
from ...instance import _Instance, Instance, InstanceArray, InstanceBundle
from ...primitives import PrimitiveCall
from ...bundle import BundleInstance
from ...instantiable import Instantiable
from ..elaboratable import Elaboratable


# Union of entry-types in the elaboration stack
ElabStackEntry = Union[Module, Instance, InstanceArray, InstanceBundle]


@datatype(config=AllowArbConfig)
class ClassLevelCache:
    """# Class-Level Cache for ElabPasses"""

    # Note Modules hash *by identity*, so each instance of `Module`,
    # regardless of the similarity of their content, gets its own entry in these sets.
    done: Set[Module] = field(default_factory=set)
    pending: Set[Module] = field(default_factory=set)


class ElabPass:
    """
    # Base ElabPass Class

    Defines the hierarchy-traversing methods used by each "elaborator pass" sub-class.

    For most passes, elaborating a `Module` at a time is the primary activity.
    To offload the need for each sub-class to cache `Module` definitions,
    and to traverse their internal contents, the base class defines `elaborate_module_base`, which:
    * Checks for the module result being pre-cached
    * Pre-order traverses its instances, arrays, and bundle instances
    * Calls the per-pass `elaborate_module`
    * Adds the result to the `modules` cache.
    This requires that sub-classes also carefully audit when they call their own
    `elaborate_module` method. Generally, they should not, and should always call
    `elaborate_module_base` instead.

    `ElabPass` and `elaborate_module_base` also manage the `ClassLevelCache` for each pass-class.
    Subclasses should not need to know there is such a cache, and should not need to access it directly.
    """

    # The class-level cache.
    # Each sub-class gets its own.
    # The base-class does not have one, should be the only `ElabPass` with `CLASS_CACHE=None`.
    CLASS_LEVEL_CACHE: Optional[ClassLevelCache] = None

    def __init_subclass__(cls) -> None:
        # Create a new `ClassLevelCache` for each subclass
        cls.CLASS_LEVEL_CACHE = ClassLevelCache()

    @classmethod
    def elaborate(cls, tops: List[Elaboratable]) -> List[Elaboratable]:
        """Elaboration entry-point. Elaborate the top-level objects."""
        return cls(tops).elaborate_tops()

    def __init__(self, tops: List[Elaboratable]):
        self.tops = tops
        self.stack: List[ElabStackEntry] = list()

    def elaborate_tops(self) -> List[Elaboratable]:
        """Elaborate our top nodes"""
        if not isinstance(self.tops, List):
            self.fail(f"Invalid Top for Elaboration: {self.tops} must be a list")

        # The base-class ElabPass, and most (perhaps all) sub-classes, return the `self.tops` list unmodified,
        # while modifiying its elements inline.
        for t in self.tops:
            self.elaborate_module_base(t)  # Note `_base` here!
        return self.tops

    def elaborate_module_base(self, module: Module) -> Module:
        """# Base-Case `Module` Elaboration

        For most passes, elaborating a `Module` at a time is the primary activity.
        To offload the need for each sub-class to cache `Module` definitions,
        and to traverse their internal contents, the base class defines `elaborate_module_base`, which:

        * Checks for the module result being pre-cached
        * Pre-order traverses its instances, arrays, and bundle instances
        * Calls the per-pass `elaborate_module`
        * Adds the result to the `modules` cache.

        This requires that sub-classes also carefully audit when they call their own
        `elaborate_module` method. Generally, they should not, and should always call
        `elaborate_module_base` instead.
        """

        # Check if this has already been elaborated by this pass/ class
        if module in self.CLASS_LEVEL_CACHE.done:
            return module

        # Add `module` to our elab stack.
        # This is helpful even if (especially if) we find it's a circular dependency next.
        self.stack.append(module)

        # Check for circular dependencies
        if module in self.CLASS_LEVEL_CACHE.pending:
            msg = f"Invalid self referencing/ circular dependency in `{module}`"
            return self.fail(msg)
        self.CLASS_LEVEL_CACHE.pending.add(module)

        # Depth-first traverse instances, ensuring their targets are defined
        for inst in module.instances.values():
            self.elaborate_instance_base(inst)
        for arr in module.instarrays.values():
            self.elaborate_instance_base(arr)
        for instbundle in module.instbundles.values():
            self.elaborate_instance_base(instbundle)

        # Traverse Bundle instances
        for bundle in module.bundles.values():
            self.elaborate_bundle_instance(bundle)

        # Run the pass-specific `elaborate_module`
        result = self.elaborate_module(module)

        # Pop the hierarchy-stack and return it
        self.stack.pop()
        self.CLASS_LEVEL_CACHE.pending.remove(module)
        self.CLASS_LEVEL_CACHE.done.add(module)
        return result

    def elaborate_module(self, module: Module) -> Module:
        """Elaborate a Module. Returns the Module unmodified by default."""
        return module

    def elaborate_external_module(self, call: ExternalModuleCall) -> ExternalModuleCall:
        """Elaborate an ExternalModuleCall. Returns the Call unmodified by default."""
        return call

    def elaborate_primitive_call(self, call: PrimitiveCall) -> PrimitiveCall:
        """Elaborate a PrimitiveCall. Returns the Call unmodified by default."""
        return call

    def elaborate_bundle_instance(self, inst: BundleInstance) -> BundleInstance:
        """Elaborate an BundleInstance"""
        # Annotate each BundleInstance so that its pre-elaboration `PortRef` magic is disabled.
        inst._elaborated = True
        return inst

    def elaborate_instance_base(self, inst: _Instance) -> Instantiable:
        """Elaborate a Module Instance, Array or Bundle thereof."""

        self.stack.append(inst)
        # Turn off `PortRef` magic
        inst._elaborated = True
        # And visit the Instance's target
        rv = self.elaborate_instantiable(inst.of)
        self.stack.pop()
        return rv

    def elaborate_instantiable(self, of: Instantiable) -> Instantiable:
        # This version of `elaborate_instantiable` is the "post-generators" version used by *most* passes.
        # The Generator-elaborator is different, and overrides it.
        if not of:
            self.fail(f"Error elaborating undefined Instance-target {of}")
        if isinstance(of, Module):
            return self.elaborate_module_base(of)  # Note `_base` here!
        if isinstance(of, PrimitiveCall):
            return self.elaborate_primitive_call(of)
        if isinstance(of, ExternalModuleCall):
            return self.elaborate_external_module(of)
        raise TypeError

    def flatname(
        self, segments: List[str], *, avoid: Optional[Dict] = None, maxlen: int = 511
    ) -> str:
        """Create a attribute-name merging string-list `segments`, while avoiding all keys in dictionary `avoid`.
        Commonly re-used while flattening  nested objects and while creating explicit attributes from implicit ones.
        Raises a `RunTimeError` if no such name can be found of length less than `maxlen`.
        The default max-length is 511 characters, a value representative of typical limits in target EDA formats.
        """

        if avoid is None:
            avoid = {}

        # The default format and result is of the form "seg0_seg1".
        # If that is in the avoid-keys, append underscores until it's not, or fails.
        name = "_".join(segments)
        while True:
            if len(name) > maxlen:
                msg = f"Could not generate a flattened name for {segments}: (trying {name})"
                self.fail(msg)
            if name not in avoid:  # Done!
                break
            name += "_"  # Collision; append underscore
        return name

    def fail(self, msg: str):
        """Error helper, adding stack and state info to an error"""
        lines = self.format_hier_path()
        msg = "Elaboration Error at hierarchical path: \n" + "".join(lines) + msg
        # This also serves as a very helpful place for a debugger breakpoint.
        raise RuntimeError(msg)

    def format_hier_path(self) -> List[str]:
        """Create a list of lines describing the current hierarchy path
        Generally intended for debug use, especially in error messages."""

        lines = []
        for s in self.stack:
            # Format: `  Module          MyModule      `
            s_name = s.name or ""
            # Filter out None-valued names, common during elaboration errors.
            line = "  " + type(s).__name__.ljust(14) + s_name.ljust(28)

            source_info = getattr(s, "_source_info", None)
            if source_info is not None:
                # The entry has source/line info. Add it to the error line.

                # For source-paths in the run-directory, print the (presumably shorter) relative-path only
                filepath = source_info.filepath

                # Path.is_relative_to() was added in Python 3.9. And writing a replacement seems harder than you'd think.
                # Use it for Python versions where it works, and skip it for 3.8.
                if (
                    sys.version_info.major > 3 or sys.version_info.minor > 8
                ) and filepath.is_relative_to(Path.cwd()):
                    filepath = filepath.relative_to(Path.cwd())
                else:  # Otherwise print the full absolute path
                    filepath = filepath.resolve()

                # Format: /path/to/file.py:123
                # Serves as a clickable link in popular IDEs.
                line += f"{str(filepath)}:{source_info.linenum}"

            line += "\n"
            lines.append(line)
        return lines



"""
# Slice and Concatenation Resolution 
"""

# Std-Lib Imports
from typing import List, Union

# Local imports
from ...module import Module
from ...signal import Signal
from ...slice import Slice
from ...concat import Concat
from ...portref import PortRef
from ...bundle import BundleRef
from ..helpers.width import width, Sliceable

# Import the base class
from .base import ElabPass


class SliceResolver(ElabPass):
    """Elaboration pass to resolve slices and concatenations to concrete signals.
    Modifies connections to any nested slices, nested concatenations, or combinations thereof.
    "Full-width" `Slice`s e.g. `sig[:]` are replaced with their parent `Signal`s.

    TODO: `Slice`s with non-unit `step` are converted to `Concat`s."""

    def elaborate_module(self, module: Module) -> Module:
        # All arrays must be flattened before getting here, or fail
        if module.instarrays:
            msg = f"Error attempting to resolve slices on {module} - "
            msg += f"still has Instance Arrays {module.instarrays}"
            raise RuntimeError(msg)

        # Then do the real work, updating any necessary connections on each instance
        for inst in module.instances.values():
            # Update all `Slice` and `Concat` valued connections to remove nested `Slice`s
            for portname, conn in inst.conns.items():
                if isinstance(conn, (Slice, Concat)):
                    resolved = _resolve_sliceable(conn)
                    inst.connect(portname, resolved)
                # All other connection-types (Signals, Interfaces) are fine

        return module


def _resolve_sliceable(conn: Sliceable) -> Sliceable:
    """Resolve a `Sliceable` to flat-concatenation-amenable elements."""
    if isinstance(conn, Signal):
        return conn  # Nothing to do
    if isinstance(conn, Slice):
        return _resolve_slice(conn)
    if isinstance(conn, Concat):
        return _resolve_concat(conn)
    if isinstance(conn, (PortRef, BundleRef)):
        return _resolve_ref(conn)
    raise TypeError(f"Invalid attempt to resolve slicing on {conn}")


def _list_slice(slize: Slice) -> List[Slice]:
    """Internal recursive helper for `resolve_slice`.
    Returns a list of Slices in which each element has a concrete Signal for its parent."""

    # Resolve "full-width" slices to their parent Signals
    if width(slize) == width(slize.parent):
        # Return a single-element list, after resolution
        return [_resolve_sliceable(slize.parent)]

    if isinstance(slize.parent, Signal):
        return [slize]  # Already all good! Just make a one-element list.

    # Do some actual work. Recursively peel off a bit at a time.
    if width(slize) == 1:
        # Base case: slice is one-bit wide. Reach into the parent signal and grab that bit.

        if isinstance(slize.parent, Slice):
            parent = slize.parent  # Note this is also a Slice
            return _list_slice(parent.parent[parent.bot + slize.bot])

        if isinstance(slize.parent, Concat):
            idx = 0  # Find the `part` including our index
            for part in slize.parent.parts:
                if width(part) + idx > slize.bot:
                    return _list_slice(part[slize.bot - idx])
                idx += width(part)
            msg = f"Slice {slize} is out of bounds of Concat {slize.parent}"
            raise RuntimeError(msg)

        raise TypeError(f"Invalid attempt to resolve slicing on {slize}")

    # Otherwise recurse in something like a "cons" pattern, splitting between the first bit and the rest.
    step = slize.step
    if step < 0:  # Negative step, begin from `top`
        first = _list_slice(slize.parent[slize.top])
        rest = slize.parent[slize.top + step : slize.bot : step]
        rest = _list_slice(rest)

    else:  # Positive step, begin from `bot`
        first = _list_slice(slize.parent[slize.bot])
        rest = _list_slice(slize.parent[slize.bot + step : slize.top : step])

    return first + rest


def _resolve_slice(slize: Slice) -> Sliceable:
    """Resolve a `Slice` to one or more with "concrete" `Signal`s as parents.

    Slices of other Slices and Slices of Concats are both valid design-time constructions.
    For example:
    ```python
    h.Concat(sig1, sig2, sig3)[1] # Slice of a Concat
    sig4[0:2][1] # Slice of a Slice
    ```

    While these may not frequently be created by designers, they are (at least) often created by array broadcasting.
    As some point their parents must be resolved to their original Signals, at minimum before export-level name resolution.

    Resolving Concatenations can generally resolve to more than one Slice, as in:
    ```python
    h.Concat(sig1[0], sig2[0], sig3[0])[0:1] # Requires slices of `sig1` and `sig2`
    ```
    Such cases create and return a Concatenation."""

    # Break out the slice elements in a list
    ls = _list_slice(slize)
    # And convert to either a single element or Concat
    if len(ls) == 1:  # Resolved to single Slice
        return ls[0]
    elif len(ls) > 1:  # Multiple parts required - concatenate them
        return Concat(*ls)

    raise RuntimeError(f"Error resolving Slice {slize}")


def _resolve_concat(conc: Concat) -> Concat:
    """Resolve a Concatenation into (a) concrete Signals and (b) Slices of concrete Signals.
    Removes nested concatenations and resolves slices along the way."""

    if not len(conc.parts):
        raise RuntimeError("Concatenation with no parts")

    if all(_flat_concatable(p) for p in conc.parts):
        return Concat(*[_resolve_sliceable(p) for p in conc.parts])

    if isinstance(conc.parts[0], Concat):
        # Recursively cover the first element, and all others
        first = _resolve_concat(conc.parts[0])
        rest = _resolve_concat(Concat(*conc.parts[1:]))
        return Concat(*(first.parts + rest.parts))

    if isinstance(conc.parts[0], Slice):
        # Resolve everything within the Slice to a list of concrete-Signal slices
        first = _resolve_slice(conc.parts[0])
        # Pass everything else recursively back to this method
        rest = _resolve_concat(Concat(*conc.parts[1:]))
        # And concatenate the two
        return Concat(*(first + rest.parts))

    # Otherwise peel off as many Signals and concrete-Signal Slices as we can
    for idx in range(len(conc.parts)):
        if _flat_concatable(conc.parts[idx]):
            continue
        # Hit our first "compound" entry. Split the list here.
        first = conc.parts[:idx]
        rest = _resolve_concat(Concat(*conc.parts[idx:]))
        return Concat(*(first + rest.parts))

    raise RuntimeError("Unable to resolve concatenation")


def _resolve_ref(ref: Union[PortRef, BundleRef]) -> Sliceable:
    """Resolve a reference to a Port or Bundle"""
    if ref.resolved is None:
        raise RuntimeError(f"Unresolved reference {ref}")
    return _resolve_sliceable(ref.resolved)


def _flat_concatable(s: Sliceable) -> bool:
    """Boolean indication of whether `s` is suitable for flattened Concatenations.
    Such objects must be either:
    * (a) A Signal, or
    * (b) A Slice into a Signal
    Notable exceptions include Concats and nested Slices of Concats and other Slices."""
    return isinstance(s, Signal) or (
        isinstance(s, Slice) and isinstance(s.parent, Signal)
    )



"""
# `Diff` Differential Bundle and `Pair` Instance Pair 
"""

# Std-Lib Imports
from enum import Enum, auto

# Local imports
from .signal import Signals
from .instance import InstanceBundleType
from .bundle import bundle, AnonymousBundle
from .role import RoleSet


class SourceSink(Enum):
    SOURCE = auto()
    SINK = auto()


SourceSink = RoleSet.from_enum(SourceSink)


@bundle
class Diff:
    """# Differential Bundle"""

    Roles = SourceSink
    p, n = Signals(2, src=Roles.SOURCE, dest=Roles.SINK)


def inverse(d: Diff) -> AnonymousBundle:
    """Create a Bundle with the same signals as `d`, but with `p` and `n` reversed."""
    return AnonymousBundle(p=d.n, n=d.p)


# Instance Pair Class
Pair = InstanceBundleType(
    name="Pair",
    bundle=Diff,
    doc=f"""
        # Differential Instance Pair 

        Two identical instances of the same `Module` which can be connected to either `Diff` signal bundles or scalar `Signal`s.  
        The rules for connecting to the ports of `Pair`s are: 

        * When connecting to a `Diff`, the `p` and `n` signals are connected to the `p` and `n` instances, respectively. 
        * When connecting to a `Signal`, the `p` and `n` instances are connected in parallel to that signal. 
        * All other connections are disallowed. 

        Example: 
    
        ```python 
        import hdl21 as h

        @h.module 
        class DiffResistors:
            # Two identical, differential resistors to VSS
            VSS = h.Port()
            diff = h.Diff()
            # The `Pair` instance
            rs = h.Pair(MyResistor)(p=diff, n=VSS)
        ```
    """,
)


__all__ = ["Diff", "inverse", "Pair"]



"""
# Hdl21 Slices

References by numeric index into Signals and other Connectable types. 
"""

from typing import Optional, Union, Any, Set
from weakref import WeakSet

# Local imports
from .datatype import datatype, AllowArbConfig
from .connect import connectable
from .sliceable import sliceable, is_sliceable
from .concat import concatable


@sliceable
@concatable
@connectable
@datatype(config=AllowArbConfig)
class Slice:
    """
    # Slice

    Subset of the indices of a parent `Connectable`,
    commonly Signals, Concatenations, and other Slices.
    Typically produced via square-bracket indexing into said `Connectable`s.

    Hdl21 slices are indexed "Python style", in the senses that:
    * Negative indices are supported, and count from the "end" of the Signal.
    * Slice-ranges such as `sig[0:2]` are supported, and *inclusive* of the start, while *exclusive* of the end index.
    * Negative-range slices such as `sig[2:0:-1]`, again *inclusive* of the start, *exclusive* of the end index, and *reversed*.

    Popular HDLs commonly use different signal-indexing conventions.
    Hdl21's own primary exchange format (in ProtoBuf) does as well,
    eschewing adopting inclusive-endpoints and negative-indexing.
    """

    # Parent Connectable.
    # Really of union-type `Sliceable`, which is more painful to type-check statically,
    # although the constructor does it procedurally.
    parent: Any
    # Python index, i.e. that passed to square brackets
    index: Union[int, slice]

    def __post_init__(self):
        if not is_sliceable(self.parent):
            raise TypeError(f"{self.parent} is not Sliceable")
        self._connected_ports: Set["PortRef"] = set()
        self._inner: Optional[SliceInner] = None
        self._slices: WeakSet[Slice] = set()
        self._concats: WeakSet["Concat"] = set()

    @property
    def top(self) -> int:
        return _get_inner(self).top

    @property
    def bot(self) -> int:
        return _get_inner(self).bot

    @property
    def step(self) -> int:
        return _get_inner(self).step

    @property
    def width(self) -> int:
        return _get_inner(self).width

    def __repr__(self):
        return f"Slice(parent={self.parent}, index={self.index})"

    def __eq__(self, other) -> bool:
        # Identity is equality
        return other is self

    def __hash__(self) -> bool:
        # Identity is equality
        return hash(id(self))


@datatype
class SliceInner:
    """Inner, private, resolved attributes of a `Slice`.
    Designed solely to be created by `_slice_inner` and stored as the `Slice._inner` field."""

    top: int  # Top index (exclusive)
    bot: int  # Bottom index (inclusive)
    step: int  # Python-convention step size
    width: int


def _slice_inner(slize: Slice) -> SliceInner:
    """Calculate the inner resolved fields for `slize`"""

    parent = slize.parent
    index = slize.index

    if isinstance(index, int):
        if index >= parent.width:
            raise ValueError(f"Out-of-bounds index {index} into {parent}")
        if index < 0:
            index += parent.width
        return SliceInner(top=index + 1, bot=index, step=1, width=1)

    if isinstance(index, slice):
        # Note these `slice` attributes are descriptor-things, and they get weird, fast.
        # Extracting their three index fields the most-hardest way via `__getattribute__` seems to work cleanest.
        start = slice.__getattribute__(index, "start")
        stop = slice.__getattribute__(index, "stop")
        step = slice.__getattribute__(index, "step")

        step = 1 if step is None else step
        if step == 0:
            raise ValueError(f"slice step cannot be zero")
        elif step < 0:
            # Here `top` gets a "+1" since `start` is *inclusive*, while `bot` gets "+1" as `stop` is *exclusive*.
            top = (
                parent.width
                if start is None
                else start + 1
                if start >= 0
                else parent.width + start + 1
            )
            bot = (
                0
                if stop is None
                else stop + 1
                if stop >= 0
                else parent.width + stop + 1
            )
            # Align bot with the step
            bot += (top - bot) % abs(step)
        else:
            # Here `start` and `stop` match `top` and `bot`'s inclusive/exclusivity.
            # No need to add any offsets.
            top = (
                parent.width
                if stop is None
                else stop
                if stop >= 0
                else parent.width + stop
            )
            bot = 0 if start is None else start if start >= 0 else parent.width + start
            # Align top with the step
            top -= (top - bot) % step

        width = (top - bot) // step

        # Create and return our Slice. More checks are done in its constructor.
        return SliceInner(top=top, bot=bot, step=step, width=width)

    # Shouldn't be reachable, but blow up if we (somehow) get here.
    raise TypeError("Internal Error: Slice index should be an int or (python) slice")


def _get_inner(slice: Slice) -> SliceInner:
    """Get a slice's `SliceInner`, calculating it inline if necessary"""
    if slice._inner is None:
        slice._inner = _slice_inner(slice)
    return slice._inner



"""
# Generators

Functions that return Modules. 
Including facilities for embedding them in a hardware hierarchy. 
"""

# Std-Lib imports
import inspect
from dataclasses import field
from typing import Callable, Any, Optional, Dict, Set, List, Type

# Local imports
from .datatype import datatype, AllowArbConfig
from .params import _unique_name, hasparams, isparamclass
from .default import Default
from .call import param_call
from .source_info import SourceInfo, source_info
from .module import Module


@datatype
class Generator:
    """
    # Generator Object

    Typically created by the `@hdl.generator` decorator.
    Stores a function-object and parameters-type, along with some auxiliary data.
    """

    func: Callable[[Any], Module]
    paramtype: Type[object]

    # Optional fields
    # Boolean indication of whether to cache calls, or return a new Module with each call.
    enable_cache: bool = True

    def __post_init__(self):
        if not isparamclass(self.paramtype):
            raise TypeError(
                f"Invalid paramtype {self.paramtype} for Generator {self.func}"
            )
        self._source_info: Optional[SourceInfo] = source_info(get_pymodule=True)

    def __call__(self, arg: Any = Default, **kwargs: Dict) -> Module:
        params = param_call(callee=self, arg=arg, **kwargs)
        call = GeneratorCall(gen=self, params=params)
        return run(call)

    def __repr__(self) -> str:
        return f"Generator(name={self.name})"

    def __eq__(self, other) -> bool:
        # Identity is equality
        return other is self

    def __hash__(self) -> bool:
        # Identity is equality
        return hash(id(self))

    @property
    def name(self) -> str:
        """Generator Name
        Equal to its callable-function's name."""
        return self.func.__name__

    @property
    def Params(self) -> type:
        """Type-style alias for the parameter-type."""
        return self.paramtype


def validate(f: Callable, **kwargs: Dict) -> Generator:
    """# Validate function `f` to be a `Generator`."""

    def invalid(s: str) -> RuntimeError:
        # If anything went wrong, write a generic message plus a more specific one.
        msg = f"Invalid call signature for generator function `{f.__name__}`.\n"
        msg += f"Generators must be of the form `def GenFunc(params: Params) -> Module`, where `Params` is a `paramclass`. \n"
        msg += s
        return RuntimeError(msg)

    if not callable(f):
        raise RuntimeError(f"Invalid `@generator` application to non-callable {f}")

    # Grab, parse, and validate its call-signature
    sig = inspect.signature(f)

    args = list(sig.parameters.values())
    if len(args) != 1:
        msg = f"Generator `{f.__name__}()` must have a single argument which is a `paramclass`."
        raise invalid(msg)

    # Extract the parameters-argument type
    paramtype = args[0].annotation
    if not isparamclass(paramtype):
        msg = f"First argument must be a paramclass, not `{paramtype}`."
        raise invalid(msg)

    # Validate the return type is `Module`
    rt = sig.return_annotation
    if rt is not Module:
        msg = f"`{f.__name__}()` must return (and must be annotated to return) a Module, not `{rt}`."
        raise invalid(msg)

    # And return the `Generator` object
    return Generator(func=f, paramtype=paramtype, **kwargs)


def generator(f: Optional[Callable] = None, **kwargs: Dict) -> Generator:
    """Decorator for Generator Functions"""

    # Accept either with or without parenthetical modifier arguments.
    inner = lambda func: validate(func, **kwargs)
    if f is None:
        return inner  # Called with parentheses
    return inner(f)  # Called without parentheses


@datatype
class GeneratorCall:
    """
    # Generator Call

    Combination of a `Generator` function-wrapper and its parameters.
    Serves as the key in the `GeneratorCache`.
    """

    gen: Generator
    params: Any

    def __eq__(self, other) -> bool:
        """
        # Generator-Call Equality
        Requires:
        - *Identity* between generators, and
        - *Equality* between parameter-values.
        """
        return self.gen is other.gen and self.params == other.params

    def __hash__(self):
        """
        # Generator-Call Hashing
        Consistent with `__eq__` above, uses:
        - *Identity* of its generator, and
        - *Value* of its parameters.
        The two are joined for hashing as a two-element tuple.
        """
        return hash((id(self.gen), self.params))


def run(call: GeneratorCall) -> Module:
    """Run Generator-function-call `call`. Returns the generated Module."""

    # First and foremost - caching.
    # See if we've already run this generator-parameters combo.
    the_cache = Generator.Cache
    if call.gen.enable_cache:
        cached_result = the_cache.done.get(call, None)
        if cached_result is not None:
            return cached_result

    # Add to the call stack.
    # This is helpful even if (especially if) we find it's a circular dependency next.
    the_cache.stack.append(call)

    if call.gen.enable_cache:
        # Check for circular dependencies.
        # Note this uses a hash-set of `GeneratorCall`s, so only hashable ones get checked.
        if call in the_cache.pending:
            msg = f"Invalid self referencing/ circular dependency in `{call}`"
            raise RuntimeError(msg)
        the_cache.pending.add(call)

    # Check that the call has a valid instance of the generator's parameter-class
    if not isinstance(call.params, call.gen.Params):
        msg = f"Invalid Generator Call {call}: {call.gen.Params} instance required, got {call.params}"
        raise RuntimeError(msg)

    # The main event: Run the generator-function
    m = call.gen.func(call.params)

    if not isinstance(m, Module):
        msg = f"Generator {call.gen} returned {m}, must return `Module`."
        raise RuntimeError(msg)

    # Give the result a reference back to the generating `Call`
    m._generated_by = call

    # Module naming
    # If the Module that comes back is anonymous, start by giving it a name equal to the Generator's
    if m.name is None:
        m.name = call.gen.name

    # If it has a nonzero number of parameters, add a unique suffix per its parameter-values
    if hasparams(call.gen.Params):
        m.name += "(" + _unique_name(call.params) + ")"

    # Store the result in our cache, and on the Call.
    the_cache.stack.pop()
    if call.gen.enable_cache:
        the_cache.pending.remove(call)
        the_cache.done[call] = m

    # And return the generated Module
    return m


@datatype(config=AllowArbConfig)
class GeneratorCache:
    """
    # Generator "Cache" (and then some)

    - Cache GeneratorCalls to their (Module) results
      - (Yes, GeneratorCalls can be hashed.)
    - Track pending and completed GeneratorCalls
    """

    done: Dict[GeneratorCall, Module] = field(default_factory=dict)
    pending: Set[GeneratorCall] = field(default_factory=set)
    stack: List[GeneratorCall] = field(default_factory=list)

    def reset(self):
        """# Reset
        Clear everything in the Generator cache."""
        self.done.clear()
        self.pending.clear()
        self.stack.clear()


# Create the `Cache`, and affix it to the `generator` function and class
generator.cache = Generator.Cache = GeneratorCache()

# Star-exports
__all__ = ["Generator", "generator"]



import pytest as pt
from decimal import Decimal
from pydantic import ValidationError
from pydantic.dataclasses import dataclass

import hdl21 as h
from hdl21.datatype import OurBaseConfig


def test_decimal():
    """This isn't a test of Hdl21 so much as a demo and reminder of how
    the standard library's `Decimal` works, particularly in conjunction with
    our widely-used `pydantic.dataclasses`.

    In short: `pydantic` converts `Decimal` valued fields to strings before
    passing them to the `Decimal` constructor.
    https://pydantic-docs.helpmanual.io/usage/types/"""

    # Decimal-only
    assert Decimal(11) == Decimal(11)
    assert Decimal(11.1) == Decimal(11.1)
    assert Decimal("11.1") == Decimal("11.1")
    # Note these are *not* equal, due to floating-point precision
    assert Decimal(11.1) != Decimal("11.1")

    # Create a dataclass with a `Decimal` field
    @dataclass
    class HasDecimal:
        d: Decimal

    assert HasDecimal(11).d == Decimal(11)
    assert HasDecimal("11.1").d == Decimal("11.1")
    # These last two are the trick.
    assert HasDecimal(11.1).d == Decimal("11.1")
    assert HasDecimal(11.1).d != Decimal(11.1)


def test_prefix_numbers():
    """Check that we get the numeric types we expect in each `Prefixed`"""

    x = h.Prefixed.new(11, h.Prefix.FEMTO)
    assert isinstance(x.number, Decimal)
    assert x.number == 11

    x = h.Prefixed.new(11.1, h.Prefix.FEMTO)
    assert isinstance(x.number, Decimal)
    assert x.number == Decimal("11.1")

    x = h.Prefixed.new(Decimal("11.11"), h.Prefix.FEMTO)
    assert isinstance(x.number, Decimal)
    assert x.number == Decimal("11.11")


def test_prefix_shortname():
    assert h.prefix.y == h.Prefix.YOCTO
    assert h.prefix.z == h.Prefix.ZEPTO
    assert h.prefix.a == h.Prefix.ATTO
    assert h.prefix.f == h.Prefix.FEMTO
    assert h.prefix.p == h.Prefix.PICO
    assert h.prefix.n == h.Prefix.NANO
    assert h.prefix.µ == h.Prefix.MICRO
    assert h.prefix.m == h.Prefix.MILLI
    assert h.prefix.K == h.Prefix.KILO
    assert h.prefix.M == h.Prefix.MEGA
    assert h.prefix.G == h.Prefix.GIGA
    assert h.prefix.T == h.Prefix.TERA
    assert h.prefix.P == h.Prefix.PETA
    assert h.prefix.E == h.Prefix.EXA
    assert h.prefix.Z == h.Prefix.ZETTA
    assert h.prefix.Y == h.Prefix.YOTTA


def test_prefix_from_exp():
    """Test creating `Prefix` from an integer exponent."""

    assert h.Prefix.from_exp(-24) == h.Prefix.YOCTO
    assert h.Prefix.from_exp(-21) == h.Prefix.ZEPTO
    assert h.Prefix.from_exp(-18) == h.Prefix.ATTO
    assert h.Prefix.from_exp(-15) == h.Prefix.FEMTO
    assert h.Prefix.from_exp(-12) == h.Prefix.PICO
    assert h.Prefix.from_exp(-9) == h.Prefix.NANO
    assert h.Prefix.from_exp(-6) == h.Prefix.MICRO
    assert h.Prefix.from_exp(-3) == h.Prefix.MILLI
    assert h.Prefix.from_exp(-2) == h.Prefix.CENTI
    assert h.Prefix.from_exp(-1) == h.Prefix.DECI
    assert h.Prefix.from_exp(0) == h.Prefix.UNIT
    assert h.Prefix.from_exp(1) == h.Prefix.DECA
    assert h.Prefix.from_exp(2) == h.Prefix.HECTO
    assert h.Prefix.from_exp(3) == h.Prefix.KILO
    assert h.Prefix.from_exp(6) == h.Prefix.MEGA
    assert h.Prefix.from_exp(9) == h.Prefix.GIGA
    assert h.Prefix.from_exp(12) == h.Prefix.TERA
    assert h.Prefix.from_exp(15) == h.Prefix.PETA
    assert h.Prefix.from_exp(18) == h.Prefix.EXA
    assert h.Prefix.from_exp(21) == h.Prefix.ZETTA
    assert h.Prefix.from_exp(24) == h.Prefix.YOTTA


def test_prefix_mul():
    """Test `Prefix` multiplication."""
    from hdl21.prefix import µ, M, e

    assert 5 * µ == 5 * h.Prefix.MICRO
    assert 5 * µ == h.Prefixed.new(5, h.Prefix.MICRO)
    assert µ * M == e(0)


def test_prefix_div():
    """Test `Prefix` division"""
    from hdl21.prefix import µ, M, e

    assert M / µ == e(12)
    assert µ / M == e(-12)


def test_prefix_pow():
    """Test `Prefix` exponentiation"""
    from hdl21.prefix import µ, e

    assert µ**2 == e(-12)
    assert µ**0.5 == e(-3)
    assert µ**-1 == e(6)


def test_prefixed_mul():
    """Test `Prefixed` Multiplication"""
    from hdl21.prefix import e

    assert (1 * e(1)) * (1 * e(2)) == 1 * e(3)
    assert 2 * (1 * e(2)) == 2 * e(2)
    assert (1 * e(2)) * 2 == 2 * e(2)
    assert (1 * e(2)) * 2.0 == 2.0 * e(2)


def test_prefixed_div():
    """Test `Prefixed` True Division"""
    from hdl21.prefix import e

    assert (1 * e(0)) / 2 == 0.5 * e(0)
    assert (1 * e(0)) / (2 * e(0)) == 0.5 * e(0)
    assert 2 / (1 * e(0)) == 2 * e(0)
    assert 6 / (3 * e(0)) == 2 * e(0)


def test_prefixed_pow():
    """Test `Prefixed` Power"""
    from hdl21.prefix import e

    assert (1 * e(1)) ** 2 == 1 * e(2)
    assert (2 * e(-2)) ** 2 == 4 * e(-4)

    # Test with __rpow__
    assert 2 ** (2 * e(0)) == 4 * e(0)
    assert 2.0 ** (2 * e(0)) == 4 * e(0)
    assert Decimal(2) ** (2 * e(0)) == 4 * e(0)


def test_prefixed_addition():
    """Test `Prefixed Addition"""
    from hdl21.prefix import e

    assert (1 * e(0)) + (1 * e(0)) == 2 * e(0)
    assert (1 * e(0)) + (1 * e(-2)) == 101 * e(-2)
    assert (1 * e(2)) + (1 * e(0)) == 101 * e(0)

    # Precision-error tests
    assert (1 * e(0)) + (1 * e(0)) == 0.2 * e(1)
    assert (1 * e(0)) + (1 * e(-2)) == 1.01 * e(0)
    assert (1 * e(2)) + (1 * e(0)) == 1.01 * e(2)

    # Addition with other types:
    assert (1 * e(0)) + 1 == 2 * e(0)
    assert 1 + (1 * e(0)) == 2 * e(0)
    assert (1 * e(0)) + 1.0 == 2.0 * e(0)
    assert 1.0 + (1 * e(0)) == 2.0 * e(0)
    assert (1 * e(0)) + Decimal(1) == 2.0 * e(0)
    assert Decimal(1) + (1 * e(0)) == 2.0 * e(0)


def test_prefixed_subtraction():
    """Test `Prefixed` Subtraction"""
    from hdl21.prefix import e

    assert (1 * e(0)) - (1 * e(0)) == 0 * e(0)
    assert (1 * e(0)) - (1 * e(-2)) == 99 * e(-2)
    assert (1 * e(2)) - (1 * e(0)) == 99 * e(0)

    # Precision-error tests
    assert (1 * e(0)) - (1 * e(0)) == 0.0 * e(1)
    assert (1 * e(0)) - (1 * e(-2)) == 0.99 * e(0)
    assert (1 * e(2)) - (1 * e(0)) == 0.99 * e(2)

    # Subtraction with other types:
    assert (1 * e(0)) - 1 == 0 * e(0)
    assert 1 - (1 * e(0)) == 0 * e(0)
    assert (1 * e(0)) - 1.0 == 0.0 * e(0)
    assert 1.0 - (1 * e(0)) == 0.0 * e(0)
    assert (1 * e(0)) - Decimal(1) == 0.0 * e(0)
    assert Decimal(1) - (1 * e(0)) == 0.0 * e(0)


def test_e():
    """Test the `e` shorthand notation for exponential creation"""
    from hdl21.prefix import e

    assert e(-24).symbol == h.Prefix.YOCTO
    assert e(-21).symbol == h.Prefix.ZEPTO
    assert e(-18).symbol == h.Prefix.ATTO
    assert e(-15).symbol == h.Prefix.FEMTO
    assert e(-12).symbol == h.Prefix.PICO
    assert e(-9).symbol == h.Prefix.NANO
    assert e(-6).symbol == h.Prefix.MICRO
    assert e(-3).symbol == h.Prefix.MILLI
    assert e(-2).symbol == h.Prefix.CENTI
    assert e(-1).symbol == h.Prefix.DECI
    assert e(1).symbol == h.Prefix.DECA
    assert e(2).symbol == h.Prefix.HECTO
    assert e(3).symbol == h.Prefix.KILO
    assert e(6).symbol == h.Prefix.MEGA
    assert e(9).symbol == h.Prefix.GIGA
    assert e(12).symbol == h.Prefix.TERA
    assert e(15).symbol == h.Prefix.PETA
    assert e(18).symbol == h.Prefix.EXA
    assert e(21).symbol == h.Prefix.ZETTA
    assert e(24).symbol == h.Prefix.YOTTA


def test_e_mult():
    """Test multiplying by the `e` function results, e.g. `11 * e(-9)"""
    from hdl21.prefix import e

    assert 11 * e(-9) == h.Prefixed.new(11, h.Prefix.NANO)
    assert 11 * e(1.5) * e(4.5) == 11 * e(6)
    assert 11 * e(1) * e(-1) == 11 * e(0)
    assert 11 * e(1) * e(-2) * e(3) == 11 * e(2)
    assert 1 * e(-0.5) * e(1) == 1 * e(0.5)

    # Precision-error tests
    assert 1 * e(-3) * e(3.3) == 1 * e(0.3)
    assert 1 * e(-0.75) * e(0.5) == 1 * e(-0.25)
    assert 1 * e(-0.123) * e(0.003) * e(0.1) == 1 * e(-0.02)
    assert 11 * e(0.123) * e(0.123) * e(0.123) == 11 * e(0.369)
    assert 1 * e(0.123) * e(-0.123) == 1 * e(0)

    # Test __mul__
    assert e(-9) * 11 == h.Prefixed.new(11, h.Prefix.NANO)
    assert e(1.5) * e(4.5) * 11 == 11 * e(6)
    assert e(1) * e(-1) * 11 == 11 * e(0)
    assert e(1) * e(-2) * e(3) * 11 == 11 * e(2)
    assert e(-0.5) * e(1) * 1 == 1 * e(0.5)


def test_e_pow():
    """Test raising Prefixed numbers to powers"""
    from hdl21.prefix import e

    assert (1 * e(1)) ** 2 == 1 * e(2)
    assert (2 * e(-2)) ** 2 == 4 * e(-4)

    # Precision-error tests
    assert (3 * e(4)) ** 0.25, Decimal(3**0.25) * e(1)


def test_e_div():
    """Test dividing Prefixed numbers by numbers"""
    from hdl21.prefix import e

    assert e(3) / e(2) == e(1)
    assert (1 * e(3)) / (1 * e(2)) == 1 * e(1)

    # Test scalar division
    assert (1 * e(2)) / 2 == 0.5 * e(2) == 5 * e(1)

    # Precision-error tests
    assert (-1 * e(-2)) / (-1 * e(-2.3)) == 1 * e(0.3)
    assert (2 * e(2.9)) / (4 * e(0.4)) == 0.5 * e(2.5)


def test_prefix_scaling():
    """Test cases of `Prefixed` multiplication which do not land on other `Prefix`es, and require scaling"""
    from hdl21.prefix import e

    # Explicit scaling
    assert (Decimal("11.11") * e(2)).scale(e(0).symbol) == Decimal(1111) * e(0)
    assert (Decimal("1.11") * e(-1)).scale(e(-3).symbol) == Decimal(111) * e(-3)
    assert (Decimal(111) * e(0)).scale(e(3).symbol) == Decimal("0.111") * e(3)

    # Inline Scaling
    assert Decimal("11.11") * e(2) * e(0) == Decimal(1111) * e(0)
    assert Decimal("1.11") * e(-2) * e(-3) == Decimal("11.1") * e(-6)
    assert 111 * e(3) * e(3) == Decimal("0.111") * e(9)

    # Automatic Scaling
    assert (1000 * e(0)).scale(), 1 * e(3)
    assert (Decimal("0.001") * e(0)).scale(), 1 * e(-3)
    assert (1000 * e(3)).scale(), 1 * e(6)


def test_prefix_comparison():
    """Test cases of `Prefixed` comparison operators"""
    from hdl21.prefix import e

    # Less than
    assert 1 * e(0) < 2 * e(0)
    assert 1 * e(3) < 1 * e(6)
    assert 1 * e(-6) < 1 * e(-3)

    # Less than or equal to
    assert 1 * e(0) <= 2 * e(0)
    assert 1 * e(0) <= 1 * e(0)
    assert 1 * e(3) <= 1 * e(6)
    assert 1 * e(3) <= 1 * e(3)
    assert 1 * e(-6) <= 1 * e(-3)
    assert 1 * e(-6) <= 1 * e(-6)

    # Not Equal to
    assert 1 * e(0) != 2 * e(0)
    assert 1 * e(-12) != 0.00001 * e(-5)
    assert 1 * e(6) != 10000 * e(1)

    # Greater than
    assert 2 * e(0) > 1 * e(0)
    assert 1 * e(6) > 1 * e(3)
    assert 1 * e(-3) > 1 * e(-6)

    # Greater than or equal to
    assert 2 * e(0) >= 1 * e(0)
    assert 1 * e(0) >= 1 * e(0)
    assert 1 * e(6) >= 1 * e(3)
    assert 1 * e(3) >= 1 * e(3)
    assert 1 * e(-3) >= 1 * e(-6)
    assert 1 * e(-6) >= 1 * e(-6)

    # Equal to with special emphasis on precision error
    assert 0.123 * e(3) == 123 * e(0)
    assert 0.456 * e(-2) == 4.56 * e(-3)
    assert (2 * e(1 / 3)) ** 2 == 4 * e(2 / 3)


def test_prefixed_conversion():
    """Test types that can be converted to `Prefixed`'s internal `Decimal`."""
    from hdl21.prefix import e, y

    assert h.Prefixed(number="11.11", prefix=y) == h.Prefixed.new(
        Decimal("11.11"), prefix=y
    )
    assert h.Prefixed(number=11.11, prefix=y) == h.Prefixed.new(
        Decimal("11.11"), prefix=y
    )
    assert h.Prefixed(number=11, prefix=y) == h.Prefixed.new(Decimal(11), prefix=y)

    assert type(float(1 * e(1))) == float
    assert float(1 * e(1)) == 10.0
    assert type(int(2 * e(2))) == int
    assert int(2 * e(2)) == 200


def test_prefix_conversion():
    """Test types that can be converted to `Prefix`'s internal `Decimal`."""
    from hdl21.prefix import m, M

    assert int(m) == 0
    assert float(m) == 0.001
    assert int(M) == 1e6
    assert float(M) == 1e6


def test_unit_prefix():
    """Test the UNIT prefix"""
    assert h.Prefixed.new(11) == h.Prefixed.new(11, h.Prefix.UNIT)
    assert h.Prefixed.new("11") == h.Prefixed.new(11)
    assert h.Prefixed.new(11.0) == h.Prefixed.new(11.0)


def test_not_implemented_prefix():
    """Test all NotImplemented Returns of Prefix Class"""
    from hdl21.prefix import K

    with pt.raises(TypeError):
        assert K * [] == NotImplemented
    with pt.raises(TypeError):
        assert [] * K == NotImplemented
    with pt.raises(TypeError):
        assert K / 3 == NotImplemented
    with pt.raises(TypeError):
        assert 3 / K == NotImplemented
    with pt.raises(TypeError):
        assert K ** [] == NotImplemented


def test_not_implemented_prefixed():
    """Test all NotImplemented Returns of Prefixed Class"""
    from hdl21.prefix import e

    with pt.raises(TypeError):
        assert type((1 * e(1)) + []) == NotImplemented
    with pt.raises(TypeError):
        assert type([] + (1 * e(1))) == NotImplemented
    with pt.raises(TypeError):
        assert type((1 * e(1)) - []) == NotImplemented
    with pt.raises(TypeError):
        assert type([] - (1 * e(1))) == NotImplemented
    with pt.raises(TypeError):
        assert type([] * (2 * e(1))) == NotImplemented
    with pt.raises(TypeError):
        assert type((2 * e(2)) * []) == NotImplemented
    with pt.raises(TypeError):
        assert type((1 * e(0)) / []) == NotImplemented
    with pt.raises(TypeError):
        assert type([] / (1 * e(0))) == NotImplemented
    with pt.raises(TypeError):
        assert type((1 * e(4)) ** []) == NotImplemented
    with pt.raises(TypeError):
        assert type([] ** (1 * e(4))) == NotImplemented


def test_to_prefixed():
    from hdl21.prefix import to_prefixed

    assert to_prefixed(1) == h.Prefixed.new(1)
    assert to_prefixed(1.0) == h.Prefixed.new(1.0)
    assert to_prefixed(Decimal(1)) == h.Prefixed.new(Decimal(1))
    assert to_prefixed(1 * h.Prefix.KILO) == h.Prefixed.new(1, h.Prefix.KILO)
    # with pt.raises(RuntimeError):
    #     assert to_prefixed([]) == NotImplemented


def test_exponent_conversions():

    from hdl21.prefix import e

    assert e(1).__repr__() == "e(1)"
    assert e(1).__str__() == "10"
    assert e(1).__float__() == 10.0


def test_prefixed_conversion():

    from hdl21.prefix import n

    assert (1 * n).__repr__() == "1*NANO"
    assert (1 * n).__str__() == "1*NANO"
    assert (1 * n).__float__() == 1e-9


def test_prefixed_abs():

    from hdl21.prefix import n

    assert abs(-1 * n) == 1 * n


def test_prefixed_neg():

    from hdl21.prefix import n

    assert -(1 * n) == -1 * n


def test_prefixed_divzero():

    from hdl21.prefix import n

    assert 1 * n / 0 == float("inf")


def test_not_implemented_exponent():
    from hdl21.prefix import Exponent, K, e

    with pt.raises(TypeError):
        assert [] * Exponent(K, Decimal(0)) == NotImplemented
    with pt.raises(TypeError):
        assert Exponent(K, Decimal(0)) * [] == NotImplemented
    with pt.raises(TypeError):
        assert Exponent(K, Decimal(0)) / [] == NotImplemented
    with pt.raises(TypeError):
        assert Exponent(K, Decimal(0)) ** [] == NotImplemented
    assert e([]) == NotImplemented


@pt.mark.xfail(reason="Testing for #157 pydantic v2")
def test_prefixed_and_scalar_conversions():
    """Test inline conversions of built-in numeric types to `Prefixed` and `Scalar`."""

    @dataclass(config=OurBaseConfig)
    class P:
        x: h.Prefixed
        y: h.Scalar

    # Test with int for each
    p = P(x=1, y=1)
    assert p.x == h.Prefixed(number=Decimal(1))
    assert p.y == h.Prefixed(number=Decimal(1))

    # Test with float for each
    p = P(x=3.0, y=3.0)
    assert p.x == h.Prefixed(number=Decimal(3.0))
    assert p.y == h.Prefixed(number=Decimal(3.0))

    # Test with str for each
    p = P(x="2e-9", y="2e-9")
    assert p.x == h.Prefixed(number=Decimal("2e-9"))
    assert p.y == h.Prefixed(number=Decimal("2e-9"))

    # Test with an expression-literal for the `Scalar`
    p = P(x=11.11, y="m*x+b")
    assert p.x == h.Prefixed(number=11.11)
    assert p.y == h.Literal(text="m*x+b")

    # Test some invalid types
    with pt.raises(ValidationError):
        p = P(x=None, y=2)
    with pt.raises(ValidationError):
        p = P(x=3, y=None)



"""
# Properties

Properties are a simple key-value store for arbitrary data.
Keys must be strings. Values may be any type.

Properties serve as a handy place for storing "off-schema" data 
on HDL objects such as Signals, Modules, and Instances.
"""

from typing import Dict, Any
from dataclasses import field

from .datatype import datatype


@datatype
class Properties:
    """
    # Properties

    Properties are a simple key-value store for arbitrary data.
    Keys must be strings. Values may be any type.

    Properties serve as a handy place for storing "off-schema" data
    on HDL objects such as Signals, Modules, and Instances.
    """

    # Wrapped dictionary
    inner: Dict[str, Any] = field(default_factory=dict)

    """ 
    Primary API: `set`, `get`, `remove`, and their dict-like counterparts.
    """

    def set(self, key: str, value: Any) -> None:
        """Set property `key` to `value`.
        Raises a `TypeError` if `key` is not a string. `value` may be any type."""
        if not isinstance(key, str):
            raise TypeError(f"key must be str, not {type(key)}")
        self.inner[key] = value

    def get(self, key: str) -> Any:
        """Get property `key`.
        Raises a `TypeError` if `key` is not a string."""
        if not isinstance(key, str):
            raise TypeError(f"key must be str, not {type(key)}")
        return self.inner.get(key)

    def remove(self, key: str) -> None:
        """Remove property `key`.
        Raises a `TypeError` if `key` is not a string."""
        if not isinstance(key, str):
            raise TypeError(f"key must be str, not {type(key)}")
        del self.inner[key]

    def __getitem__(self, key: str) -> Any:
        return self.get(key)

    def __setitem__(self, key: str, value: Any) -> None:
        self.set(key, value)

    def __delitem__(self, key: str) -> None:
        self.remove(key)

    """
    Dictionary wrapper special methods
    """

    def __repr__(self) -> str:
        return f"Properties({self.inner})"

    def __contains__(self, key: str) -> bool:
        return key in self.inner

    def __len__(self) -> int:
        return len(self.inner)

    def __iter__(self):
        return iter(self.inner)

    def __eq__(self, other) -> bool:
        return self.inner == other.inner



"""
# Hdl21 Connectable Decorator & Type Union

Applied to types that are "used as connections", e.g. Signal. 
"""

# Std-Lib Imports
from typing import Any, Union, TypeVar, Type

T = TypeVar("T")


def connectable(cls: Type[T]) -> Type[T]:
    """Decorator for connectable types"""
    cls.__connectable__ = True
    return cls


def is_connectable(obj: Any) -> bool:
    """Boolean indication of connect-ability"""
    return getattr(obj, "__connectable__", False)


# Union of types using `connectable`
# For checking, `is_connectable` is preferable, but this serves as a handy shorthand for many type annotations.
Connectable = Union[
    "Signal",
    "Slice",
    "Concat",
    "NoConn",
    "PortRef",
    "BundleInstance",
    "AnonymousBundle",
    "BundleRef",
]



"""
# hdl21.module 
The `Module` module (get it?)

This module primarily defines: 
* The core circuit-reuse element `hdl21.Module` 
* The `@module` (lower-case) decorator-function, for class-syntax creation of `Module`s
"""

from inspect import isclass
from typing import Any, Optional, Union, List, Dict

# Local imports
from .source_info import source_info, SourceInfo
from .attrmagic import init
from .signal import Signal, Visibility
from .instance import (
    calls_instantiate,
    Instance,
    InstanceArray,
    InstanceBundle,
)
from .bundle import BundleInstance
from .props import Properties
from .literal import Literal

# Type-alias for HDL objects storable as `Module` attributes
ModuleAttr = Union[Signal, Instance, InstanceArray, InstanceBundle, BundleInstance]
# And "other things" that are also storable as `Module` attributes
ModuleOther = Union[Literal, Properties]


@calls_instantiate
@init
class Module:
    """# Module

    The central element of hardware re-use.
    Hdl21 `Module`s are static combinations of HDL objects such as `Signal`s, `Instance`s of other `Module`s, and `Bundle`s.

    Two primary methods (directly) create `Module`s. The `Module` (capital M) class-constructor accepts a single optional `name` field.
    This serves as the name of an initially-empty Module-definition. Attributes can then be added programmatically.
    ```
    m = Module(name="MyModule")
    m.inp = h.Input(width=5)
    m.add(h.Output(name="out", width=4))
    ```

    The `hdl21.module` (lower-case m) decorator-function is a convenient way to create a `Module` from a Python class-definition.
    Each attribute in the class-body is converted to a module attribute, and the class name is used as the module's name.
    ```
    @h.module
    class MyModule:
        inp = h.Input(width=5)
        out = h.Output(width=4)
    ```

    Hdl21 Modules *do not* contain behavior in the sense of procedural HDLs. Nor do they contain parameters.
    Parametric hardware is produced through Hdl21's `generator` facility, which defines python functions which create and return `Module`s.
    """

    def __init__(self, name: Optional[str] = None):
        if name is not None and not isinstance(name, str):
            # Something wrong with this `name`.
            # Most common case: confusion with the `module` decorator.
            if isclass(name):
                msg = f"Module name {name} is a class. Did you mean to use the `module` decorator?"
            else:
                msg = f"Invalid Module name {name}, must be a string"
            raise TypeError(msg)
        self.name = name

        self.ports: Dict[str, Signal] = dict()
        self.signals: Dict[str, Signal] = dict()
        self.instances: Dict[str, Instance] = dict()
        self.instarrays: Dict[str, InstanceArray] = dict()
        self.instbundles: Dict[str, InstanceBundle] = dict()
        self.bundles: Dict[str, BundleInstance] = dict()
        self.namespace: Dict[str, ModuleAttr] = dict()  # Combination of all these

        self.literals: List[Literal] = list()
        self.props: Properties = Properties()

        # Elaborated version of this module.
        # Set at the end of elaboration.
        # For most modules this will be `self`.
        self._elaborated: Optional[Module] = None

        # IOs as captured before bundle-flattening.
        # Bundle-valued ports are flattened into `ports` and removed from `bundles`, but need to be kept *somewhere* afterwards.
        # Set to `None` initially to indicate that it hasn't been set yet.
        self._pre_flattening_io: Optional[Dict[str, "Connectable"]] = None

        # The source `GeneratorCall`, for generated Modules.
        self._generated_by: Optional["GeneratorCall"] = None

        self._importpath = None  # Optional field set by importers
        self._source_info: Optional[SourceInfo] = source_info(get_pymodule=True)
        self._initialized = True

    """
    The public `Module` API. Just `add` and `get`. (FIXME: eventually.)
    
    `Module`s are generally used and thought of as arbitrary HDL objects. For example: 
    ```python
    m = Module(name="MyModule")
    m.i = h.Input()
    m.q = AnotherModule(i=m.i)
    ```

    To keep the space of valid names for post-fix dot access as large as possible, 
    the `Module` class namespace is kept *as small* as possible. 
    """

    def add(self, val: ModuleAttr, *, name: Optional[str] = None) -> ModuleAttr:
        """Add a Module attribute.

        `Module.add` allows for programmatic insertion of attributes whose names are not legal Python identifiers,
        such as keywords ('in', 'from') and those including invalid characters.

        The added object `val` is also provided as the return value, enabling chaining-style usages such as
        ```python
        instance.inp = MyModule.add(h.Input(name="in", width=5))
        ```
        and similar.

        Attribute naming comes from one of either:
        * `val`'s `name` attribute, if it has one, or
        * The optional `name` argument.
        Only one of the two name-sources is allowed per call.
        """

        # Check it's a valid attribute-type
        _assert_module_attr(self, val)

        # Now sort out naming. We get two name-sources:
        # (a) the function-argument `name` and (b) the value's `name` attribute.
        # One or the other (and not both) must be set.
        if name is None and val.name is None:  # Neither set, fail.
            msg = f"Anonymous attribute {val} cannot be added to Module {self.name}"
            raise RuntimeError(msg)
        if name is not None and val.name is not None:  # Both set, fail.
            msg = f"{val} with conflicting names {name} and {val.name} cannot be added to Module {self.name}"
            raise RuntimeError(msg)
        if name is not None:  # One or the other set - great.
            val.name = name

        # Now `val.name` is set appropriately.
        # Add it to our type-based containers, and return it.
        return _add(module=self, val=val)

    def get(self, name: str) -> Optional[ModuleAttr]:
        """Get module-attribute `name`. Returns `None` if not present.
        Note unlike Python built-ins such as `getattr`, `get` returns solely
        from the HDL namespace-worth of `ModuleAttr`s."""
        ns = self.__getattribute__("namespace")
        return ns.get(name, None)

    @property
    def bundle_ports(self) -> dict:
        """Port-Exposed Bundle Instances"""
        # FIXME: this shall be kicked out of the `Module` class and namespace
        return {name: bundle for name, bundle in self.bundles.items() if bundle.port}

    """
    Special Methods
    """

    def __setattr__(self, key: str, val: Any) -> None:
        """Set-attribute over-ride, organizing into type-based containers"""

        if key.startswith("_") or not getattr(self, "_initialized", False):
            # Bootstrapping phase. Pass along to "regular" setattr.
            return super().__setattr__(key, val)

        if key in _banned:
            msg = f"Error attempting to over-write protected attribute {key} of Module {self}"
            raise RuntimeError(msg)
        # Special case(s)
        if key == "name":
            return super().__setattr__(key, val)

        # Check it's a valid attribute-type
        _assert_module_attr(self, val)

        # Checks out! Name `val` and add it to our type-based containers.
        val.name = key
        _add(module=self, val=val)
        return None

    def __getattr__(self, key: str) -> Any:
        """Include our namespace-worth of HDL objects in dot-access retrievals"""
        if key.startswith("_"):
            return object.__getattribute__(self, key)
        ns = self.__getattribute__("namespace")
        if key in ns:
            return ns[key]
        return object.__getattribute__(self, key)

    def __delattr__(self, __name: str) -> None:
        """Disable attribute deletion.
        This may be enabled some day, but until unwinding any dependencies is not allowed.
        """
        msg = f"Cannot delete Module attribute {__name} of {self}"
        raise RuntimeError(msg)

    def __init_subclass__(cls, *_, **__):
        """Sub-Classing Disable-ization"""
        msg = f"Error attempting to create {cls.__name__}. Sub-Typing {cls} is not supported."
        raise RuntimeError(msg)

    def __repr__(self) -> str:
        if self.name:
            return f"Module(name={self.name})"
        return f"Module(_anon_)"

    def __eq__(self, other) -> bool:
        # Identity is equality
        return other is self

    def __hash__(self) -> bool:
        # Identity is equality
        return hash(id(self))


def module(cls: type) -> Module:
    """
    # Module Definition Decorator

    Converts a class-body full of Module-storable attributes to an `hdl21.Module`.
    Example Usage:

    ```python
    import hdl21 as h

    @h.module
    class M1:
        d = h.Port()
        e = h.Signal()

    @h.module
    class M2:
        q = h.Signal()
        i = M1(d=q)
    ```

    `hdl21.Modules` are strongly-typed containers of other hardware objects:
    * `Signals` and `Ports`
    * Instances of other `Modules`
    * Structured connections via `Bundles`

    Attempts to set module attributes of any other types will raise a `TypeError`.
    Notably this includes any behavioral elements implemented by Python functions.
    If you'd like arbitrary-executing Python-code that creates or manipulates `Modules`,
    you probably want an `hdl21.Generator` instead.

    Temporary variables can be declared inside of the `@module`-decorated class-body,
    so long as they are meet the Python convention for private data: name-prefixing with an underscore.
    These temporary variables are *not* propagated along as members of the `Module`.
    """

    if not isclass(cls):
        msg = f"Error attempting to create {cls.__name__}. @module can only decorate classes."
        raise RuntimeError(msg)

    if cls.__bases__ != (object,):
        raise RuntimeError(f"Invalid @hdl21.module inheriting from {cls.__bases__}")

    # Create the Module object
    module = Module(name=cls.__name__)

    # Any class-body content that isn't a `ModuleAttr` will be "forgotten" from the `Module` definition.
    # This can nonetheless be handy for defining intermediate values upon which the ultimate Module attributes depend.
    forgetme: List[Any] = list()

    # Take a lap through the class dictionary, type-check everything and assign relevant attributes to the bundle
    for key, val in cls.__dict__.items():
        if key in _banned:
            raise RuntimeError(f"Invalid field name {key} in Module {module.name}")
        elif _is_module_attr(val):
            setattr(module, key, val)
        else:  # Add to the forget-list
            forgetme.append(val)

    # And return the Module
    return module


""" 
(Python) Module-Level Functions

Many of which could be methods, but are generally kept here to prevent expanding the `Module` namespace.
"""


# Protected Module attribute names
_banned = [
    "ports",
    "signals",
    "instances",
    "instarrays",
    "instbundles",
    "bundles",
    "literals",
    "props",
    "namespace",
    "add",
    "get",
]


def _add(module: Module, val: ModuleAttr) -> ModuleAttr:
    """Internal `Module.add` and `Module.__setattr__` implementation.
    Primarily sort `val` into one of our type-based containers.
    Layers above `_add` must ensure that `val` has its `name` attribute before calling this method.
    """

    if module._elaborated is not None:
        raise RuntimeError(f"Cannot add {val} to {module} after elaboration.")

    # Sort out which of our type-based containers to add `val` to.
    if isinstance(val, Signal):
        if val.vis == Visibility.PORT:
            type_ctr = module.ports
        else:
            type_ctr = module.signals
    elif isinstance(val, Instance):
        type_ctr = module.instances
    elif isinstance(val, InstanceArray):
        type_ctr = module.instarrays
    elif isinstance(val, InstanceBundle):
        type_ctr = module.instbundles
    elif isinstance(val, BundleInstance):
        type_ctr = module.bundles
    else:
        # The next line *should* never be reached, as outer layers should have checked `_is_module_attr`.
        # Nonetheless gotta raise an error if we get here, somehow.
        _attr_type_error(val)

    # Add it to the module namespace, and the type-specific container
    type_ctr[val.name] = val
    module.namespace[val.name] = val

    # Give it a reference to us.
    #
    # ## Note:
    # Parent-testing *can* be done here instead of at elaboration time.
    # It's not clear that this would be a good idea though.
    # Attributes which are *copied* (for example) keep the `_parent_module` member, which is often helpful to just over-write.
    # Nonetheless if "setattr-time" failure is desired, this is where it will go:
    #
    # _parent = getattr(val, "_parent_module", None) # Note many attributes will not have this member, until this assignment.
    # if _parent is not None and _parent is not module:
    #     msg = f"{val.name} being added to {module} already has a parent-module {val._parent_module}"
    #     raise RuntimeError(msg)
    #
    # Ok, now actually give it a reference to us:
    val._parent_module = module

    # And return our newly-added attribute
    return val


def _is_module_attr(val: Any) -> bool:
    """Boolean indication of whether `val` is a valid `hdl21.Module` attribute."""
    return isinstance(val, ModuleAttr.__args__)


def _assert_module_attr(m: Module, val: Any) -> None:
    """Raise a TypeError if `val` is not a valid Module attribute."""
    if not _is_module_attr(val):
        _attr_type_error(m, val)


def _attr_type_error(m: Module, val: Any) -> None:
    """Raise a `TypeError` with debug info for invalid attribute `val`."""

    # Give more specific error-messages for our common types.
    # Especially those that are easily confused, e.g. all the `Call` types, `Module`s, and `Instance`s thereof.
    from .external_module import ExternalModule, ExternalModuleCall
    from .generator import Generator, GeneratorCall
    from .primitives import Primitive, PrimitiveCall

    if isinstance(val, (Generator, Primitive, ExternalModule)):
        msg = f"Cannot add `{type(val).__name__}` `{val.name}` to `Module` `{m.name}`. Did you mean to make an `Instance` by *calling* it - once for params and once for connections - first?"
    elif isinstance(val, Module):
        msg = f"Cannot add `{type(val).__name__}` `{val.name}` to `Module` `{m.name}`. Did you mean to make an `Instance` by *calling* to connect it first?"
    elif isinstance(val, (GeneratorCall, PrimitiveCall, ExternalModuleCall)):
        msg = f"Cannot add `{type(val).__name__}` `{val}` to `Module` `{m.name}`. Did you mean to make an `Instance` by *calling* to connect it first?"
    else:
        msg = f"Invalid Module attribute {val} of type {type(val)} for {m}. Valid `Module` attributes are of types: {list(ModuleAttr.__args__)}"
    raise TypeError(msg)


__all__ = ["module", "Module"]



# PyPi Imports
from pydantic.dataclasses import dataclass


@dataclass(frozen=True)
class Literal:
    """
    # Literal
    A thin wrapper around a string, generally to be used outside Hdl21, e.g. in external netlists.
    """

    text: str  # String literal text


__doc__ = Literal.__doc__



"""
# Orphan-Testing (Orphanage) ElabPass
"""

# Local imports
from ...module import Module, ModuleAttr
from ...instance import _Instance

# Import the base class
from .base import ElabPass


class Orphanage(ElabPass):
    """# Orphan-Checking ElabPass Pass

    Ensures each Module-attribute is "parented" by the `Module` which holds it.
    Errant cases can come up for code such as:

    ```
    m1 = h.Module(name='m1')
    m1.s = h.Signal() # Signal `s` is now "parented" by `m1`

    m2 = h.Module(name='m2')
    m2.y = m1.s # Now `s` has been "orphaned" (or perhaps "cradle-robbed") by `m2`
    ```

    This essentially boils down to the difference between Python's native reference-semantics,
    and `hdl21.Module`'s notion of "owning" its attributes.
    Note other *references* to Module attributes are allowed, such as:

    ```
    m1 = h.Module(name='m1')
    m1.s = h.Signal() # Signal `s` is now "parented" by `m1`

    my_favorite_signals = { "from_m1" : m1.s }
    ```

    Here the dictionary `my_favorite_signals` retains a reference to Signal `s`.
    This does not generate an orphan-error complaint, so long as `Module`-parent is unique and unambiguous.

    The orphan-test is very simple: each Module-attribute is annotated with a `_parent_module` member
    upon insertion into the Module namespace.
    Orphan-testing simply requires that for each attribute, this member is identical to the parent Module.
    A `RuntimeError` is raised if orphaned attributes are detected.
    Otherwise each Module is returned unchanged.
    """

    def elaborate_module(self, module: Module) -> Module:
        """Elaborate a Module"""

        # Check each attribute in the module namespace for orphanage.
        for attr in module.namespace.values():
            self.assert_parentage(module, attr)

        # Check instance connections, which are not in the module namespace.
        instlike = (
            list(module.instances.values())
            + list(module.instarrays.values())
            + list(module.instbundles.values())
        )
        for inst in instlike:
            self.check_instance(module, inst)

        # Checks out! Return the module unchanged.
        return module

    def check_instance(self, module: Module, inst: _Instance) -> None:
        """Check the connections of `inst` in parent `module`"""
        self.stack.append(inst)

        # Check each of the instance's connections
        for conn in inst.conns.values():
            self.check_connectable(module, conn)

        self.stack.pop()

    def check_connectable(self, module: Module, conn: "Connectable") -> None:
        """Check a Connectable for orphanage.
        Dispatches across connectable types, and recursively follows `conn` back to its constituent and/or parent elements."""

        from ... import (
            NoConn,
            Signal,
            Slice,
            Concat,
            BundleInstance,
            BundleRef,
            AnonymousBundle,
            PortRef,
        )

        # Check owned types first: Signals and Bundle Instances
        if isinstance(conn, (Signal, BundleInstance)):
            return self.assert_parentage(module, conn)

        if isinstance(conn, Slice):
            # Recursively check the parent-signals of slices
            return self.check_connectable(module, conn.parent)

        if isinstance(conn, Concat):
            # Check each of the concatenated signals, also recursively across inner types
            for part in conn.parts:
                self.check_connectable(module, part)
            return

        if isinstance(conn, NoConn):
            # `NoConn`s are not "parented" by anything; they are essentially exempt from this check.
            return

        if isinstance(conn, PortRef):
            # For Port references, check that `module` owns their target Instance
            return self.assert_parentage(module, conn.inst)

        if isinstance(conn, BundleRef):
            # For Bundle references, check that `module` owns their root Bundle Instance
            return self.assert_parentage(module, conn.root())

        if isinstance(conn, AnonymousBundle):
            for sig in conn._namespace.values():
                self.check_connectable(module, sig)
            return

        raise TypeError(f"Orphanage: Unhandled Connectable `{conn}`")

    def assert_parentage(self, module: Module, attr: ModuleAttr) -> None:
        """Assert that `attr` is parented by `module`, or fail."""

        if attr._parent_module is None:
            msg = f"Orphanage! Module `{module.name}` depends on orphan attribute `{attr}`! "
            msg += "Did you forget to `add()` or assign it into `{module.name}`? "
            self.fail(msg)

        if attr._parent_module is not module:
            msg = f"Orphanage: Module {module} attribute {attr} is actually owned by another Module {attr._parent_module}!"
            self.fail(msg)



"""
# Python Source Info

Track the source file and line number of user calls. 

This source-level tracking is invaluable for hierarchical hardware designs, 
especially since the Hdl21 elaboration process means default debugger errors 
do not produce a stack-trace through the *hardware*, but through the *elaboration* call-stack. 

Most of the logic to produce nice error messages using `SourceInfo`s 
is in `ElabPass`s, particularly the base-class `ElabPass`'s `fail` method. 

The mechanisms for tracking this source info have been the subject of some experimentation. 
Past methods, and associated lessons include: 

- `inspect.stack()` and `inspect.getmodule()` are REALLY slow. 
  - Versions of Hdl21 overusing these methods have observed them taking as much as 90% of elaboration time. 
  - `stack()` in particular does not abide the Python3 "everything is a generator" rule, and computes all stack-frames into a list upfront
  - We still use `getmodule()` as a basis for (Hdl21) Module naming. But beware to limit its calls, and get rid of it if we can. 
- Storing the `frame` object returned by `inspect.currentframe()` at creation time and referring back to it later, just doesn't work. 
  - The `frame` objects are mutated directly by the interpreter; the stack is lost essentially immediately. 

"""

import inspect
from pathlib import Path
from types import ModuleType, FrameType
from typing import Optional

from .datatype import datatype, AllowArbConfig


@datatype(config=AllowArbConfig)
class SourceInfo:
    """# Python Source Info"""

    filepath: Path
    linenum: int
    pymodule: Optional[ModuleType] = None


def source_info(get_pymodule: bool = False) -> Optional[SourceInfo]:
    """
    # Get Source Info

    Find the source details of the first python-stack-frame from outside `files_to_skip`.
    Returns `None` if *no* such modules are in the call-stack.

    Returns a `SourceInfo` with `filepath` and `linenum` fields,
    and additionally finds the `pymodule` python module field if requested
    via the `get_pymodule` boolean argument.
    """

    # Get or create the files-to-skip set
    global files_to_skip
    if files_to_skip is None:
        files_to_skip = get_files_to_skip()
    # Note: this file is in the `files_to_skip` list, so this function's call-stack entry will never show up.

    frame: Optional[FrameType] = inspect.currentframe()
    MAX_DEPTH = 1000  # Avoid infinite recursion

    for _ in range(MAX_DEPTH):
        if frame is None:
            return None
        if frame.f_code.co_filename not in files_to_skip:
            # We've got a hit! Return a `SourceInfo` object.

            # If requested via the `get_pymodule` flag, return the Python module.
            # Note this is generally the slow part.
            pymodule = None if not get_pymodule else inspect.getmodule(frame)

            return SourceInfo(
                filepath=Path(frame.f_code.co_filename),
                linenum=frame.f_lineno,
                pymodule=pymodule,
            )
        # And back up one frame
        frame = frame.f_back

    # If we got here without returning, we failed.
    raise RecursionError("Error finding `SourceDetail`")


# Set of files to skip
# Calculated once, after import-time, so those modules can import this one.
# files_to_skip: Optional[Set[str]] = None
files_to_skip = None


def get_files_to_skip():
    """Create the list of files to skip.
    Should only be run once, shortly after import-time."""

    from . import (
        _module_module,
        _external_module_module,
        _generator_module,
        _instance_module,
    )

    return set(
        [
            __file__,  # *This* file
            _module_module.__file__,
            _external_module_module.__file__,
            _generator_module.__file__,
            _instance_module.__file__,
        ]
    )



import hdl21 as h


def test_resolve_portref_type():
    """Test resolution of PortRef "types" to their Module-level referents"""
    from hdl21.elab.helpers.resolve_ref_types import resolve_portref_type

    @h.module
    class Inner:
        p = h.Port()

    @h.module
    class Outer:
        i = Inner()

    # Create a `PortRef`
    ref = Outer.i.p

    # All of these succeed:
    assert ref is Outer.i.p
    assert resolve_portref_type(ref) is not Outer.i.p
    assert resolve_portref_type(ref) is Inner.p


def test_portref_concat():
    """Test concatenating Port References"""

    @h.module
    class A:
        p = h.Port(width=1)

    @h.module
    class B:
        p = h.Port(width=3)

    @h.module
    class Outer:
        a = A()
        b = B(p=h.Concat(a.p, a.p, a.p))

    h.elaborate(Outer)


def test_portref_slice():
    """Test slicing Port References"""

    @h.module
    class A:
        p = h.Port(width=1)

    @h.module
    class B:
        p = h.Port(width=3)

    @h.module
    class Outer:
        b = B()
        a = A(p=b.p[2])

    h.elaborate(Outer)


def test_resolve_bundleref_type():
    """Test resolution of BundleRef "types" to their Bundle-level referents"""
    from hdl21.elab.helpers.resolve_ref_types import resolve_bundleref_type

    @h.bundle
    class B:
        s = h.Signal()

    @h.module
    class Outer:
        b = B()

    # Create a `BundleRef`
    ref = Outer.b.s

    # All of these succeed:
    assert ref is Outer.b.s
    assert resolve_bundleref_type(ref) is not Outer.b.s
    assert resolve_bundleref_type(ref) is B.s


def test_bundleref_concat():
    """Test concatenating Bundle References"""

    @h.module
    class Inner:
        p = h.Port(width=3)

    @h.bundle
    class B:
        s = h.Signal()

    @h.module
    class Outer:
        b = B()
        i = Inner(p=h.Concat(b.s, b.s, b.s))

    h.elaborate(Outer)


def test_bundleref_slice():
    """Test slicing Bundle References"""

    @h.module
    class Inner:
        p = h.Port(width=2)

    @h.bundle
    class B:
        s = h.Signal(width=8)

    @h.module
    class Outer:
        b = B()
        i = Inner(p=b.s[1:3])

    h.elaborate(Outer)


def test_anon_with_signal():
    """Test an AnonymousBundle with a Signal"""

    @h.bundle
    class B:
        s = h.Signal()

    @h.module
    class Inner:
        b = B(port=True)

    @h.module
    class Outer:
        s = h.Signal()
        i = Inner(b=h.AnonymousBundle(s=s))

    h.elaborate(Outer)


def test_anon_with_concat():
    """Test an AnonymousBundle with a concatenation"""

    @h.bundle
    class B:
        s = h.Signal(width=2)

    @h.module
    class Inner:
        b = B(port=True)

    @h.module
    class Outer:
        s = h.Signal(width=1)
        i = Inner(b=h.AnonymousBundle(s=h.Concat(s, s)))

    h.elaborate(Outer)


def test_anon_with_slice():
    """Test an AnonymousBundle with a slice"""

    @h.bundle
    class B:
        s = h.Signal(width=2)

    @h.module
    class Inner:
        b = B(port=True)

    @h.module
    class Outer:
        s = h.Signal(width=4)
        i = Inner(b=h.AnonymousBundle(s=s[0:2]))

    h.elaborate(Outer)


def test_anon_with_bundle_inst():
    """Test an AnonymousBundle referring to a BundleInstance"""

    @h.bundle
    class BI:
        x = h.Signal()

    @h.bundle
    class B:
        y = h.Signal()
        bi = BI()

    @h.module
    class Inner:
        b = B(port=True)

    @h.module
    class Outer:
        z = h.Signal()
        bi = BI()
        i = Inner(b=h.AnonymousBundle(y=z, bi=bi))

    h.elaborate(Outer)


def test_anon_with_bundleref_signal():
    """Test an AnonymousBundle referring to a sub-bundle reference"""

    @h.bundle
    class B:
        s = h.Signal()

    @h.module
    class Inner:
        b = B(port=True)

    @h.module
    class Outer:
        b = B()
        i = Inner(b=h.AnonymousBundle(s=b.s))

    h.elaborate(Outer)

    assert len(Outer.signals) == 1
    assert len(Outer.ports) == 0
    assert len(Outer.bundles) == 0
    assert len(Inner.signals) == 0
    assert len(Inner.ports) == 1
    assert len(Inner.bundles) == 0
    assert Inner.ports["b_s"]
    assert Outer.i.conns["b_s"] is Outer.signals["b_s"]


def test_anon_with_subbundle_ref():
    """Test an AnonymousBundle referring to a sub-bundle reference"""

    @h.bundle
    class BI:
        x = h.Signal()

    @h.bundle
    class B:
        y = h.Signal()
        bi = BI()

    @h.module
    class Inner:
        b = B(port=True)

    @h.module
    class Outer:
        z = h.Signal()
        b = B()
        i = Inner(b=h.AnonymousBundle(y=z, bi=b.bi))

    h.elaborate(Outer)


def test_anon_with_portref():
    """Test an AnonymousBundle referring to a PortRef"""

    @h.bundle
    class B:
        s = h.Signal()

    @h.module
    class HasBPort:
        b = B(port=True)

    @h.module
    class HasSignalPort:
        s = h.Port()

    @h.module
    class Outer:
        hass = HasSignalPort()
        hasb = HasBPort(b=h.AnonymousBundle(s=hass.s))

    h.elaborate(Outer)


def test_anon_with_anon():
    """Test an AnonymousBundle with a nested AnonymousBundle"""

    @h.bundle
    class BI:
        x = h.Signal()

    @h.bundle
    class B:
        y = h.Signal()
        bi = BI()

    @h.module
    class Inner:
        b = B(port=True)

    @h.module
    class Outer:
        a, b = h.Signals(2)
        i = Inner(b=h.AnonymousBundle(y=a, bi=h.AnonymousBundle(x=b)))

    h.elaborate(Outer)



"""
# Elaborate-able Types 

More specifically, things that can be *the top level* of an elaboration process. 
Sets the type-bounds on primary arguments to `elaborate` and related functions. 
"""

from typing import Any, List, Union

# Local imports
from ..module import Module
from ..primitives import PrimitiveCall
from ..external_module import ExternalModuleCall


# Type short-hands for elaborate-able types
# NOTE: this is identical to `Instantiable`.
# Although maybe some day it won't be?
Elaboratable = Union[Module, ExternalModuleCall, PrimitiveCall]
# (Plural Version)
Elaboratables = Union[Elaboratable, List[Elaboratable]]


def is_elaboratable(obj: Any) -> bool:
    # Function to test this, since `isinstance` doesn't work for `Union`.
    return isinstance(obj, Elaboratable.__args__)



"""
# Param-Class Argument Calls

Enables the "no paramclass constructor" style invocation for Generators, ExternalModules, and Primitives.
Creates an instance of the generator's parameter-class if keyword args were provided instead.
If both were provided, fail.
"""

from typing import Union, Any, Dict
from pydantic import ValidationError

# Local Imports
from .default import Default


# Union of types designed to support `param_call`,
# which more specifically must have a `Params` property, which is a type.
Callee = Union["Generator", "ExternalModule", "Primitive"]


def param_call(callee: Callee, arg: Any = Default, **kwargs: Dict) -> "Callee.Params":
    """
    # Param-Class Argument Calls

    Enables the "no paramclass constructor" style invocation for Generators, ExternalModules, and Primitives.
    Creates an instance of the generator's parameter-class if keyword args were provided instead.
    If both were provided, fail.
    """

    if kwargs and arg is not Default:
        msg = f"Invalid call to {callee}: either provide a single {callee.Params} instance or keyword arguments, not both."
        raise RuntimeError(msg)

    if arg is not Default:
        # Already constructed. Return as-is.
        # Note type-checking for instances of `callee.Params` *is not* done here.
        return arg

    try:
        # Try to create and return an instance of the parameter-class
        # As indicated by the `Params` property of the callee.
        return callee.Params(**kwargs)

    except (TypeError, ValidationError) as e:
        # Type validation failed; add some of our info to the error message.

        # A very common case is forgetting to apply parameters, and going straight to connections.
        # Check for this, and provide a specific error message when it happens.
        from .connect import is_connectable

        connectables = [key for key, val in kwargs.items() if is_connectable(val)]
        if connectables:
            msg = f"Invalid call to {callee} with connectables: ({', '.join(connectables)}). \n"
            msg += f"Did you forget to pass this its `params` ({callee.Params.__name__}) first?"
            raise RuntimeError(msg)

        # FIXME: add `SourceInfo`
        msg = f"Invalid call to `{callee}` with arguments `{kwargs}`: \n\n{e}"
        raise RuntimeError(msg)

    except:  # Any other Exception, raise along
        raise



""""
# Bundle Roles and Role Sets 
"""

# Std Lib Imports
from typing import Optional, List, Dict, Type, TypeVar
from dataclasses import field
from copy import copy
from enum import EnumMeta

# Local Imports
from .datatype import datatype

T = TypeVar("T")


@datatype
class Role:
    """# Bundle Role"""

    name: Optional[str] = None

    def __rmul__(self, num: int) -> List["Role"]:
        """# Right multiplication. Creates `num` copies of this Role."""
        if not isinstance(num, int):
            return NotImplemented
        return [copy(self) for _ in range(num)]


def Roles(num: int) -> List[Role]:
    """# Create `num` new Roles."""
    return [Role() for _ in range(num)]


@datatype
class RoleSet:
    """
    # Bundle Role Set

    Typically used to represent a set of `Role`s that are associated with a `Bundle`.
    """

    name: Optional[str] = None
    inner: Dict[str, Role] = field(default_factory=dict)

    @staticmethod
    def from_enum(enum: EnumMeta):
        """# Convert a standard-library `Enum` class into a `RoleSet`."""
        if not isinstance(enum, EnumMeta):
            raise TypeError(f"Expected EnumMeta, got {enum}")
        return RoleSet(
            name=enum.__name__, inner={role.name: Role(name=role.name) for role in enum}
        )

    @staticmethod
    def from_names(names: List[str]):
        """# Convert a list of role-names into a `RoleSet`."""
        return RoleSet(name=None, inner={name: Role(name=name) for name in names})

    @staticmethod
    def from_list(roles: List[Role]):
        """# Convert a list of `Role`s into a `RoleSet`."""
        return RoleSet(
            name=None, inner={role.name: Role(name=role.name) for role in roles}
        )

    @staticmethod
    def from_dict(roles: Dict[str, Role]):
        """# Convert a dict of `Role`s into a `RoleSet`."""
        return RoleSet(name=None, inner=roles)

    """
    # Special Methods 
    """

    def __getitem__(self, key: str) -> Role:
        """# Get a Role by name."""
        return self.inner[key]

    def __getattr__(self, name: str) -> Role:
        """# Get a Role by name."""
        if name.startswith("_"):
            return object.__getattribute__(self, name)
        return self.inner[name]


def roleset(cls: Type[T]) -> Type[T]:
    """# Decorator to convert an Enum class into a `RoleSet`."""

    return RoleSet.from_enum(cls)


__all__ = ["Role", "Roles", "RoleSet", "roleset"]



"""
# Bundle-Flattening ElabPass Pass 
"""

# Std-Lib Imports
import copy
from dataclasses import field
from typing import Dict, List, Union, Optional

# Local imports
from ...datatype import datatype, AllowArbConfig
from ...module import Module
from ...instance import Instance
from ... import Slice, Concat, NoConn, PortRef
from ...bundle import (
    AnonymousBundle,
    BundleInstance,
    BundleRef,
)
from ...signal import PortDir, Signal, Visibility
from ...instantiable import io
from ..helpers.resolve_ref_types import update_ref_deps

# Import the base class
from .base import ElabPass


@datatype(frozen=True)
class Path:
    """
    # Hierarchical String-Valued Path

    This contains a list which is mutable, but... don't mutate it.
    The `Path` object itself is designed to be hashed,
    particularly for by-key lookups of hierarchical Bundle attributes.
    Don't ever mutate them; create new ones.
    """

    segs: List[str]  # List of path segments

    def append(self, suffix: "Path") -> "Path":
        """Append `suffix`. Returns a *new* Path."""
        return Path(segs=self.segs + suffix.segs)

    def prepend(self, prefix: "Path") -> "Path":
        """Prepend `prefix`. Returns a *new* Path."""
        return Path(segs=prefix.segs + self.segs)

    def to_name(self) -> str:
        return "_".join(self.segs)

    def __hash__(self):
        return hash(tuple(self.segs))

    def __eq__(self, other: "Path") -> bool:
        if not isinstance(other, Path):
            return NotImplemented
        return self.segs == other.segs


@datatype(config=AllowArbConfig)
class BundleScope:
    """Scope-worth of Signals for a flattened Bundle"""

    src: Union[BundleInstance, AnonymousBundle]  # Source/ Original Bundle
    parent: Optional["BundleScope"] = None  # Parent Scope

    signals: Dict[Path, Signal] = field(default_factory=dict)
    # Flattened signals-dict, keyed by Path string.
    # Includes *all* Signals in *all* sub-scopes.

    scopes: Dict[Path, "BundleScope"] = field(default_factory=dict)
    # Hierarchical scopes
    # All Signals inside sub-scopes are references to elements in `signals`,
    # at the `path` as they'd be referred to in that Scope.

    def add_subscope(self, name: str, scope: "BundleScope"):
        """Add a sub `BundleScope`. Also add its signals to our parent dict."""

        name = Path([name])
        self.scopes[name] = scope
        scope.parent = self

        for path_suffix, sig in scope.signals.items():
            # And add them all to the root `signals` dict, with a prepended path
            path_from_self = path_suffix.prepend(name)
            if path_from_self in self.signals:
                msg = f"Error Flattening Bundles: colliding flattened Signal names for {sig} and {self.signals[path_from_self]}"
                raise RuntimeError(msg)
            self.signals[path_from_self] = sig


@datatype(config=AllowArbConfig)
class BundlePortEntry:
    """# Bundle-Port Entry in the Cache
    Hashable combination of a Module and a portname."""

    module: Module
    portname: str

    def __eq__(self, other: "BundlePortEntry") -> bool:
        if not isinstance(other, BundlePortEntry):
            return NotImplemented
        return self.module is other.module and self.portname == other.portname

    def __hash__(self) -> int:
        return hash((id(self.module), self.portname))


@datatype(config=AllowArbConfig)
class Cache:
    """
    The Bundle Flattening Cache
    Designed to operate at Module level, across multiple ElabPasss.
    """

    bundle_insts: Dict[int, BundleScope] = field(default_factory=dict)
    # BundleInstance replacements, {id(BundleInst) => BundleScope}

    anon_bundles: Dict[int, BundleScope] = field(default_factory=dict)
    # AnonymousBundle replacements, {id(AnonBundle) => BundleScope}

    flat_bundle_ports: Dict[BundlePortEntry, BundleScope] = field(default_factory=dict)
    # Replacement flattened Bundle-valued ports.
    # Keyed by Module and port-name.
    # {BundlePortEntry(module, portname) => BundleScope}


# The module-scope flattening cache
THE_CACHE = Cache()


class BundleFlattener(ElabPass):
    """Bundle-Flattening ElabPass Pass"""

    def elaborate_module(self, module: Module) -> Module:
        """Flatten Module `module`s Bundles, replacing them with newly-created Signals.
        Reconnect the flattened Signals to any Instances connected to said Bundles."""

        # Cache the state of the Module's IOs before flattening
        module._pre_flattening_io = copy.copy(io(module))

        # Remove and replace each `BundleInstance` from the Module
        while module.bundles:
            name, bundle_inst = module.bundles.popitem()
            module.namespace.pop(name)
            self.replace_bundle_inst(module, bundle_inst)

        # Go through each Instance, replacing `AnonymousBundle`s with their referents
        # Note `Module`s do not store `AnonymousBundle`s directly, so we don't have a quicker way
        # to find them all than traversing the connections to each instance and array.
        for inst in instances_and_arrays(module):
            anon_conns = {
                portname: conn
                for portname, conn in inst.conns.items()
                if isinstance(conn, AnonymousBundle)
            }
            for portname, anon_bundle in anon_conns.items():
                self.replace_anon_bundle_conn(
                    inst=inst, portname=portname, anon=anon_bundle
                )

        return module

    def replace_bundle_inst(self, module: Module, bundle_inst: BundleInstance):
        """Replace a `BundleInstance`, flattening its Signals into `module`'s namespace,
        and replacing all of its Instance connections with their flattened replacements.
        """

        # Check we haven't (somehow) already replaced it
        if id(bundle_inst) in THE_CACHE.bundle_insts:
            msg = f"Bundle Instance {bundle_inst} in Module {module} flattened more than once. Was it actually part of another Module?"
            self.fail(msg)

        # Flatten it
        flat = self.flatten_bundle_inst(bundle_inst, path=Path([]))

        # Add each flattened Signal. Note flattened Signals are modified in-place.
        for pathstr, sig in flat.signals.items():
            # Rename the signal, prepending the bundle-instance's name
            sig.name = self.flatname(
                segments=[bundle_inst.name, pathstr.to_name()],
                avoid=module.namespace,
            )
            # And add it to the Module namespace
            module.add(sig)

        # Store the result in our caches
        THE_CACHE.bundle_insts[id(bundle_inst)] = flat
        if bundle_inst.port:
            entry = BundlePortEntry(module, bundle_inst.name)
            THE_CACHE.flat_bundle_ports[entry] = flat

        # Replace connections to any connected instances
        for portref in list(bundle_inst._connected_ports):
            self.replace_bundle_conn(
                inst=portref.inst, portname=portref.portname, flat=flat
            )

        # Kick off recursive `Ref` replacement for any (potential nested)
        # references `bundle_inst` has given out.
        self.resolve_bundlerefs(bundle_inst)

    def resolve_bundlerefs(self, hasrefs: Union[BundleInstance, BundleRef]) -> None:
        """Resolve all BundleRefs that `hasrefs` has given out."""
        for bref in hasrefs.refs_to_me.values():
            # Recursively get references it has handed out
            self.resolve_bundlerefs(bref)
            # And resolve `bref` itself
            self.resolve_bundleref(bref)

    def replace_bundle_conn(self, inst: Instance, portname: str, flat: BundleScope):
        """
        Replace a connection to a `BundleInstance` with a connections to the Signals of a `BundleScope`.
        At this point, if we started with a circuit like:

        ```python
        @h.bundle
        class B:
            x, y, z = h.Signals(3)

        @h.module
        class Inner:
            inner_b = B(port=True)

        @h.module
        class Outer:
            outer_b = B()
            inner = Inner(b=outer_b)
        ```
        At the point of getting here to reconnect the port `inner.b`,
        the `Inner` module has been flattened, and looks more like

        ```
        @h.module
        class Inner:
            inner_b_x, inner_b_y, inner_b_z = h.Ports(3)
        ```

        The bundle instances in `Outer` have also been flattened.
        It now has signals named `outer_b_x`, `outer_b_y`, and `outer_b_z`,
        which are stored in a `BundleScope` keyed by `x`, `y`, and `z`.
        After this function it will (or should) look like:

        ```
        @h.module
        class Outer:
            outer_b_z, outer_b_y, outer_b_z  = h.Signals(3)
            inner = Inner(inner_b_x=outer_b_x, inner_b_y=outer_b_y, inner_b_z=outer_b_z)
        ```

        The primary mental gymnastics here are in naming,
        particularly between the Instance and instantiating module.
        """

        entry = BundlePortEntry(inst.of, portname)
        flat_bundle_port = THE_CACHE.flat_bundle_ports.get(entry, None)
        if flat_bundle_port is None:
            msg = f"Invalid Port Connection to {portname} on Instance {inst}"
            self.fail(msg)

        # Disconnect the old hierarchical Bundle port
        inst.disconnect(portname)

        # Replace the connection to each flattened Signal
        for path, flat_port in flat_bundle_port.signals.items():
            # Note in analogy to the commentary above -
            # * `path` will be relative to the Bundle definition
            #   * In the example, its values will be `x`, `y`, and `z`.
            # * `flat_port.name` will be the instance's new port names
            #   * In the example, its values will be `inner_b_x`, `inner_b_y`, and `inner_b_z`.
            # * Neither takes on the values `outer_b_x`, `outer_b_y`, and `outer_b_z`.
            #   * These would be `flat.signals[path].name`
            if path not in flat.signals:
                msg = f"Missing connection to `{path}` "
                msg += f"in Connection to `{portname}` on Instance `{inst.name}`. "
                msg += f"Has Signals `{list(flat.signals.keys())}`, "
                msg += f"but no `{path}`."
                self.fail(msg)
            inst.connect(flat_port.name, flat.signals[path])

    def flatten_bundle_inst(
        self, bundle_inst: BundleInstance, path: Path
    ) -> BundleScope:
        """# Flatten a bundle instance"""
        # Kick off recursive flattening
        return self.flatten_bundle_inst_helper(
            bundle_inst=bundle_inst,
            path=path,
            is_this_top_level=True,
            is_port=bundle_inst.port,
            flip_state=bundle_inst.flipped,
        )

    def flatten_bundle_inst_helper(
        self,
        bundle_inst: BundleInstance,
        path: Path,
        is_this_top_level: bool,
        is_port: bool,
        flip_state: bool,
    ) -> BundleScope:
        """Recursive inner implementation of `flatten_bundle_inst`."""

        bundle_def = bundle_inst.of
        scope = BundleScope(src=bundle_inst)

        # Copy each scalar signal, retaining its original name as the key in `scope.signals`
        for sig in bundle_def.signals.values():
            signal_path = Path([sig.name])
            if signal_path in scope.signals:
                self.fail(f"Doubly defined Signal {sig} in {bundle_inst}")
            newsig = copy.deepcopy(sig)
            newsig.name = signal_path.to_name()

            # Sort out the new Signal's visibility and direction
            if is_port:  # "Port-ness" applies to the entire top-level bundle instance
                vis_ = Visibility.PORT

                # If the Signal is declared as a port, look at its direction and our flip state
                if newsig.vis == Visibility.PORT:
                    if flip_state:  # Flip direction
                        dir_ = newsig.direction.flipped()
                    else:  # Not flipped; leave unchanged
                        dir_ = newsig.direction

                # If the Signal is not a port, look at its source and dest roles.
                elif bundle_inst.role is None:
                    dir_ = PortDir.NONE
                elif bundle_inst.role == newsig.src:
                    dir_ = PortDir.OUTPUT
                elif bundle_inst.role == newsig.dest:
                    dir_ = PortDir.INPUT
                else:  # Some other Role
                    dir_ = PortDir.NONE
            else:
                vis_ = Visibility.INTERNAL
                dir_ = PortDir.NONE

            # Apply all these attributes to our new Signal
            newsig.vis = vis_
            newsig.direction = dir_
            scope.signals[signal_path] = newsig

        # And recursively do this to all sub-bundle instances, adding them to `scopes` along the way.
        for sub_bundle_inst in bundle_def.bundles.values():
            subpath = path.append(Path([sub_bundle_inst.name]))
            sub_flip_state = (
                flip_state if not sub_bundle_inst.flipped else not flip_state
            )
            subscope = self.flatten_bundle_inst_helper(
                bundle_inst=sub_bundle_inst,
                path=subpath,
                is_this_top_level=False,
                is_port=is_port,  # Port-ness is defined from the top-level BundleInstance
                flip_state=sub_flip_state,  # Flip-state can be inverted at each level
            )
            scope.add_subscope(name=sub_bundle_inst.name, scope=subscope)

        return scope

    def replace_anon_bundle_conn(
        self, inst: Instance, portname: str, anon: AnonymousBundle
    ):
        """Replace an `AnonymousBundle` connection with its (already) flattened referents."""

        anon_scope: BundleScope = self.flatten_anonymous_bundle(anon)
        self.replace_bundle_conn(inst, portname, anon_scope)

    def flatten_anonymous_bundle(self, anon: AnonymousBundle) -> BundleScope:
        """Flatten an `AnonymousBundle`.
        Differs from bundle-class instances in that each attribute of anonymous-bundles
        are generally "references", owned by something else, commonly a Module."""

        if id(anon) in THE_CACHE.anon_bundles:
            return THE_CACHE.anon_bundles[id(anon)]

        scope = BundleScope(src=anon)

        for name, attr in anon._namespace.items():
            # First resolve any references
            # Note at this point in elaboration, these Anon-Bundles are the sole remaining place `PortRef`s can hide.
            # They are also the last place where `BundleRef`s will be resolved,
            # although the others just have been, earlier in this elaborator pass.
            if isinstance(attr, (BundleRef, PortRef)):
                attr = self.resolve_bundleref(attr)

            if isinstance(attr, NoConn):  # Invalid
                self.fail(f"Invalid AnonymousBundle NoConn attribute {attr} in {anon}")

            elif isinstance(attr, (Signal, Slice, Concat)):
                # "Scalar" signals get added to the result scope
                scope.signals[Path([name])] = attr

            # Finally handle nested, Bundle-like types, which produce sub-scopes
            elif isinstance(attr, BundleInstance):
                # BundleInstances have all been visited by now, and hence better be in the cache
                flat_inst = THE_CACHE.bundle_insts.get(id(attr), None)
                if flat_inst is None:
                    self.fail(f"Invalid AnonymousBundle attribute {attr}")
                scope.add_subscope(name, flat_inst)

            elif isinstance(attr, AnonymousBundle):
                subscope = self.flatten_anonymous_bundle(attr)
                scope.add_subscope(name, subscope)

            elif isinstance(attr, BundleScope):
                scope.add_subscope(name, attr)
            else:  # Shouldn't be reachable
                raise TypeError(f"Invalid AnonBundle attribute {attr}")

        THE_CACHE.anon_bundles[id(anon)] = scope
        return scope

    def resolve_bundleref(self, bref: BundleRef) -> Union[Signal, BundleScope]:
        """Resolve a bundle-reference to a Signal or Flattened Bundle thereof."""

        if bref.resolved is not None:
            return bref.resolved  # Already done

        # Get its path and root BundleInstance
        path: List[str] = bref.path()
        root: BundleInstance = bref.root()

        # Get the flattened version of the root BundleInstance
        flat_root = THE_CACHE.bundle_insts.get(id(root), None)
        if flat_root is None:
            msg = f"Invalid BundleRef to {bref.parent}"
            self.fail(msg)

        bref.resolved = resolved = self.resolve_path(flat_root, Path(path))

        if isinstance(resolved, BundleScope):
            for connected_port in list(bref._connected_ports):
                self.replace_bundle_conn(
                    inst=connected_port.inst,
                    portname=connected_port.portname,
                    flat=resolved,
                )
            return resolved

        if isinstance(resolved, Signal):
            update_ref_deps(bref, resolved)
            return resolved

        return self.fail(f"BundleRef {bref} resolved to invalid {resolved}")

    def resolve_path(
        self, scope: BundleScope, path: Path
    ) -> Union[Signal, BundleScope]:
        """Resolve `path` in the flattend `scope`, producing either a Signal or nested BundleScope."""

        ns = scope
        for seg in path.segs:
            seg = Path([seg])
            if seg in ns.signals:
                ns = ns.signals[seg]
            elif seg in ns.scopes:
                ns = ns.scopes[seg]
            else:
                pathstr = ".".join(path.segs)
                if isinstance(scope.src, BundleInstance):
                    msg = f"Cannot resolve path `{pathstr}` in `{scope.src.name}`"
                else:
                    msg = f"Cannot resolve path `{pathstr}` in `{scope}`"
                return self.fail(msg)
        return ns


def instances_and_arrays(module: Module) -> List[Instance]:
    """Get a list of `module`'s instances and instance arrays."""
    return list(module.instances.values()) + list(module.instarrays.values())


__all__ = ["BundleFlattener"]



import importlib
from pathlib import Path
from types import FrameType, ModuleType
from hdl21.source_info import source_info, SourceInfo


def test_source_info():
    """Test getting `SourceInfo` for the current call-stack.
    Note the `linenum` assertions here generally change if this file changes.
    This file is kept short on purpose!"""

    call = source_info(get_pymodule=True)

    assert isinstance(call, SourceInfo)
    assert call.filepath == Path(__file__)
    assert call.linenum == 12
    assert isinstance(call.pymodule, ModuleType)

    call = source_info(get_pymodule=False)

    assert isinstance(call, SourceInfo)
    assert call.filepath == Path(__file__)
    assert call.linenum == 19
    assert call.pymodule is None



"""
# Concatenation-ability Decorator 

Solely a marker used by `Concat` for validity checking. 
"""

# Local imports
from .connect import is_connectable
from typing import TypeVar, Type

T = TypeVar("T")


def concatable(cls: Type[T]) -> Type[T]:
    """Decorator for `Concat`-compatible types."""
    if not is_connectable(cls):
        raise TypeError(f"{cls} is not connectable")
    # Just adds a "marker trait" to the class
    cls.__concatable__ = True
    return cls


def is_concatable(obj: object) -> type:
    return getattr(obj, "__concatable__", False)



""" 
# Hdl21 Built-In Generators 
# Unit Tests 
"""

import hdl21 as h
from hdl21.generator import GeneratorCall


def test_mos_generator():
    """Initial test of built-in series-Mos generator"""

    Mos = h.generators.MosStack
    m = Mos(Mos.Params(nser=2))
    call = m._generated_by

    assert isinstance(call, GeneratorCall)
    assert isinstance(call.params, Mos.Params)

    m = h.elaborate(m)

    assert isinstance(m, h.Module)
    assert len(m.ports) == 4
    assert m.ports.keys() == h.primitives.Mos.ports.keys()
    assert len(m.instances) == 2
    assert "units_0" in m.instances
    assert "units_1" in m.instances
    assert isinstance(m.instances["units_0"].of, h.PrimitiveCall)
    assert isinstance(m.instances["units_1"].of, h.PrimitiveCall)
    assert m.instances["units_0"].of.params == h.primitives.MosParams()
    assert m.instances["units_1"].of.params == h.primitives.MosParams()

    assert len(m.signals) == 1
    assert "i" in m.signals


def test_series_generator():
    """# Test `h.generators.Series`"""

    from hdl21.generators import Series

    @h.module
    class M:  # Unit cell
        a, b, c, d, e, f, g = h.Ports(7)

    params = Series.Params(unit=M, nser=2, conns=["a", "b"])
    m = Series(params)
    call = m._generated_by

    assert isinstance(call, GeneratorCall)
    assert isinstance(call.params, Series.Params)

    m = h.elaborate(m)

    assert isinstance(m, h.Module)
    assert len(m.ports) == 7
    assert m.ports.keys() == M.ports.keys()
    assert len(m.instances) == 2
    assert "units_0" in m.instances
    assert "units_1" in m.instances

    assert len(m.signals) == 1
    assert "i" in m.signals

    for inst in m.instances.values():
        assert inst.of is M



""" 
# Hdl21 Built-In Generators Library 
"""

from copy import deepcopy
from typing import Tuple, Union

# This is about the one place within the library that we use the global, named import `hdl21 as h`,
# largely so that this module is a bit more copy-paste-edit-able.
import hdl21 as h


SeriesConn = Union[h.Signal, str]
SeriesConns = Tuple[SeriesConn, SeriesConn]


@h.paramclass
class SeriesParams:
    """# Series Generator Parameters"""

    # Required
    unit = h.Param(dtype=h.Instantiable, desc="Unit cell")
    conns = h.Param(dtype=SeriesConns, desc="Series ports (or names)")
    # Optional
    nser = h.Param(dtype=int, desc="Number in series", default=1)


@h.generator
def Series(params: SeriesParams) -> h.Module:
    """
    # Series Generator

    Arrays `params.nser` series-stacked Instances of unit-cell `params.unit`.
    The generated `Module` includes the same ports as `unit`.
    The two series-connected ports of `unit` are specified by parameter two-tuple `conns`.
    All other ports of `unit` are wired in parallel, and exposed as ports of the generated module.
    """

    if params.nser < 1:
        raise ValueError(f"Invalid Series() generator with nser={params.nser}")
    if params.nser == 1:
        return Wrapper(params.unit)  # Easy mode

    # Initialize our stack-module
    m = h.Module()

    # Copy the unit-cell ports
    for p in params.unit.ports.values():
        m.add(deepcopy(p))

    # Divy up the ports by series vs parallel connections
    series_conns = _seriesconns(m, params.conns)
    par_ports = [port for port in m.ports.values() if port not in series_conns]
    unit_conns = {port.name: port for port in par_ports}

    # Create the internal series-connected signals, and concatenate them with the series ports
    i = m.add(h.Signal(name="i", width=params.nser - 1))
    unit_conns[series_conns[0].name] = h.Concat(series_conns[0], i)
    unit_conns[series_conns[1].name] = h.Concat(i, series_conns[1])

    # Create an array of unit instances
    m.add(params.nser * params.unit(**unit_conns), name="units")

    # And return the module
    return m


def _seriesconns(m: h.Module, conns: SeriesConns) -> Tuple[h.Signal, h.Signal]:
    # Extract the signals specified by `SeriesConns`
    return (_seriesconn(m, conns[0]), _seriesconn(m, conns[1]))


def _seriesconn(m: h.Module, conn: SeriesConn) -> h.Signal:
    # Extract the signal specified by a `SeriesConn`
    if isinstance(conn, h.Signal):
        rv = m.ports.get(conn.name, None)
    elif isinstance(conn, str):
        rv = m.ports.get(conn, None)
    else:
        raise TypeError(f"Series: invalid series port {conn}")
    # Check that we got something, and that it's a Signal (not, e.g., a Bundle).
    if rv is None or not isinstance(rv, h.Signal):
        raise ValueError(f"Series: invalid series port {conn}")
    return rv


@h.paramclass
class MosStackParams:
    """# Mos Stack Parameters"""

    # All optional
    # Equal to `SeriesParams`, with:
    # - `conns` fixed to source/ drain
    # - `unit` has a default, the built-in `h.primitives.Mos`
    unit = h.Param(
        dtype=h.Instantiable, desc="Unit Mos cell", default_factory=h.primitives.Mos
    )
    nser = h.Param(dtype=int, desc="Number in series", default=1)


@h.generator
def MosStack(params: MosStackParams) -> h.Module:
    """# Mos Series-Stack Generator"""
    # Use the `Series` generator, with series-connections between drain and source.
    return Series(unit=params.unit, nser=params.nser, conns=("d", "s"))


def Wrapper(m: h.Instantiable) -> h.Module:
    """
    # Module Wrapper Creator

    Adds a `Module` hierarchy layer around argument-module (or other instantiable) `m`.
    Creates an `Instance` of `m` and clones its ports, connecting each.

    Note: `Wrapper` is generally aways more helpful when the returned `Module` is modified after the fact.
    Hence while `Wrapper` is a *function that returns a `Module`*, it is *not* an `hdl21.Generator`,
    which cache and unique-name their results.
    Callers of `Wrapper` are therefore responsible for considerations such as unique naming.
    """

    from .instantiable import io

    # Initialize our wrapper-module
    wrapper = h.Module(name=f"{m.name}Wrapper")

    # Copy the inner-cell ports
    # Note this also serves as the connections-dict to the inner instance
    wrapper_io = {p.name: wrapper.add(deepcopy(p)) for p in io(m).values()}

    # Create the inner instance
    wrapper.add(h.Instance(name="inner", of=m)(**wrapper_io))

    # And return the wrapper
    return wrapper


@h.paramclass
class AcDc:
    ac = h.Param(dtype=h.Prefixed, desc="AC Voltage", default=0)
    dc = h.Param(dtype=h.Prefixed, desc="DC Voltage", default=0)


@h.paramclass
class CmDmGenParams:
    cm = h.Param(dtype=AcDc, desc="Common-Mode Voltage", default_factory=AcDc)
    dm = h.Param(dtype=AcDc, desc="Differential Voltage", default_factory=AcDc)


@h.generator
def CmDmGen(p: CmDmGenParams) -> h.Module:
    """# Common & Differential Mode Generator"""

    @h.module
    class CmDmGen:
        # IO
        diff = h.Diff(port=True, role=h.Diff.Roles.SOURCE)
        VSS = h.Port()
        # Implementation
        vc = h.Vdc(dc=p.cm.dc, ac=p.cm.ac)(n=VSS)
        vp = h.Vdc(dc=p.dm.dc, ac=+p.dm.ac / 2)(p=diff.p, n=vc.p)
        vn = h.Vdc(dc=p.dm.dc, ac=-p.dm.ac / 2)(p=diff.n, n=vc.p)

    return CmDmGen


@h.generator
def Balun(_: h.HasNoParams) -> h.Module:
    """# Balun Generator"""

    @h.module
    class Balun:
        # IO
        vic = h.Diff(port=True, role=h.Diff.Roles.SINK)
        vid = h.Diff(port=True, role=h.Diff.Roles.SINK)
        vod = h.Diff(port=True, role=h.Diff.Roles.SOURCE)
        VSS = h.Ground()

        # Implementation
        voc = h.Signal()
        ec = h.Vcvs(gain=1)(p=voc, n=VSS, cp=vic.p, cn=vic.n)
        ep = h.Vcvs(gain=+500 * h.prefix.MILLI)(p=vod.p, n=voc, cp=vid.p, cn=vid.n)
        en = h.Vcvs(gain=-500 * h.prefix.MILLI)(p=vod.n, n=voc, cp=vid.p, cn=vid.n)

    return Balun



""" 
# Hdl21 Sample PDK

A non-physical, 110% made-up process technology, designed solely for demonstrating the `hdl21.pdk` interface. 
The Hdl21 sample PDK is integrated in the Hdl21 package, and is therefore available to every installation of Hdl21. 

It includes: 
* Nmos and Pmos transistor `ExternalModule`s
* Similar transistor modules, defined as SPICE primitive-MOS models
* Comilation from `hdl21.primitives` to the PDK modules 
* An `Install` type, demonstrating typical usage of out-of-Python PDK installation data 
* A built-in model file, compatible with all supported simulators

Note that unlike most Hdl21 PDKs, the content of `sample_pdk.install` - the model file - 
*is* built into the source tree, and is configured at import-time. 
For most PDKs in which this data is distributed separately, 
a site-specific `sitepdks` module customarily configures the `install` variable.

"""

# Std-Lib Imports
import copy
from pathlib import Path

# PyPi Imports
from pydantic.dataclasses import dataclass

# Project Imports
import hdl21 as h
from hdl21.prefix import µ
from hdl21.primitives import Mos, MosType, MosParams as PrimMosParams
from hdl21.pdk import PdkInstallation
from vlsirtools import SpiceType


@dataclass
class Install(PdkInstallation):
    models: Path  # Path to the models file


@h.paramclass
class SamplePdkMosParams:
    """Sample PDK MOS Transistor Parameters"""

    w = h.Param(dtype=h.Scalar, desc="Width in resolution units", default=1 * µ)
    l = h.Param(dtype=h.Scalar, desc="Length in resolution units", default=1 * µ)
    nf = h.Param(dtype=h.Scalar, desc="Number of parallel fingers", default=1)
    m = h.Param(dtype=h.Scalar, desc="Number of parallel fingers", default=1)

    def __post_init__(self):
        """Value Checks"""
        if self.w <= 0:
            raise ValueError(f"MosParams with invalid width {self.w}")
        if self.l <= 0:
            raise ValueError(f"MosParams with invalid length {self.l}")
        if self.nf <= 0:
            msg = f"MosParams with invalid number parallel fingers {self.nf}"
            raise ValueError(msg)


# `Pmos` and `Nmos` external modules, each with similar parameter-spaces to `hdl21.primitives.Mos`.
Pmos = h.ExternalModule(
    domain="sample_pdk",
    name="pmos",
    desc=f"Sample PDK Pmos",
    port_list=copy.deepcopy(Mos.port_list),
    paramtype=SamplePdkMosParams,
)
Nmos = h.ExternalModule(
    domain="sample_pdk",
    name="nmos",
    desc=f"Sample PDK Nmos",
    port_list=copy.deepcopy(Mos.port_list),
    paramtype=SamplePdkMosParams,
)

# `PmosModel` and `NmosModel` also external modules, referring directly to the SPICE "primitive MOS" M-elements.
PmosModel = h.ExternalModule(
    domain="sample_pdk",
    name="pmos_model",
    desc=f"Sample PDK Pmos (Model M-Element)",
    port_list=copy.deepcopy(Mos.port_list),
    paramtype=SamplePdkMosParams,
    spicetype=SpiceType.MOS,
)
NmosModel = h.ExternalModule(
    domain="sample_pdk",
    name="nmos_model",
    desc=f"Sample PDK Nmos (Model M-Element)",
    port_list=copy.deepcopy(Mos.port_list),
    paramtype=SamplePdkMosParams,
    spicetype=SpiceType.MOS,
)


class SamplePdkWalker(h.HierarchyWalker):
    """Hierarchical Walker, converting `h.Primitive` instances to process-defined `ExternalModule`s."""

    def __init__(self):
        super().__init__()
        # Keep a cache of `ExternalModuleCall`s to avoid creating duplicates
        self.mos_modcalls = dict()

    def visit_primitive_call(self, call: h.PrimitiveCall) -> h.Instantiable:
        """Replace instances of `h.primitive.Mos` with our `ExternalModule`s"""
        # Replace transistors
        if call.prim is h.primitives.Mos:
            return self.mos_module_call(call.params)
        # Return everything else as-is
        return call

    def mos_module(self, params: PrimMosParams) -> h.ExternalModule:
        """Retrieve or create an `ExternalModule` for a MOS of parameters `params`."""
        if params.tp == MosType.PMOS:
            return Pmos
        return Nmos

    def mos_params(self, params: PrimMosParams) -> SamplePdkMosParams:
        """Convert generic primitive `MosParams` into PDK-specific `SamplePdkMosParams`"""
        # FIXME: map parameters using the `default` field directly; needs debug
        w = params.w or 1 * µ
        l = params.l or 1 * µ
        m = params.mult or 1
        nf = params.nf or 1
        return SamplePdkMosParams(w=w, l=l, m=m, nf=nf)

    def mos_module_call(self, params: PrimMosParams) -> h.ExternalModuleCall:
        """Retrieve or create a `Call` for MOS parameters `params`."""
        # First check our cache
        if params in self.mos_modcalls:
            return self.mos_modcalls[params]

        # Not found; create a new `ExternalModuleCall`.
        # First retrieve the `ExternalModule`.
        mod = self.mos_module(params)

        # Convert to `mod`s parameter-space
        modparams = self.mos_params(params)

        # Combine the two into a call, cache and return it
        modcall = mod(modparams)
        self.mos_modcalls[params] = modcall
        return modcall


def compile(src: h.Elaboratables) -> None:
    """Compile `src` to the Sample technology"""
    return SamplePdkWalker.walk(src)



"""
# VLSIR ProtoBuf Schema Export

Primary entrypoint `to_proto` turns a set of elaborate-able `Elaboratables`, 
and turns them into a `vlsir.circuit.Package`. 

Most export-machinery is accomplished by primary implementation-class `ProtoExporter`, 
which particularly handles hierarchically-referential types such as `Module`, `Instance`, and `ExternalModule`. 

Exports of simpler "scalar"(-ish) types such as `Signal`s and `Param`s are provided as 
free-standing functions to enable use elsewhere. 
"""

from decimal import Decimal
from dataclasses import fields
from enum import Enum
from typing import Optional, List, Union, Dict, Any

# Local imports
# Proto-definitions
import vlsir
import vlsir.circuit_pb2 as vckt

# HDL
from ..datatype import datatype, AllowArbConfig
from ..params import isparamclass
from ..prefix import Prefix, Prefixed
from ..scalar import Scalar
from ..literal import Literal
from ..elab import Elaboratables, elaborate
from ..module import Module
from ..qualname import qualname as module_qualname
from ..external_module import ExternalModule, ExternalModuleCall
from ..instance import Instance
from ..signal import Signal, Port, PortDir
from ..slice import Slice
from ..concat import Concat
from ..primitives import (
    PrimitiveCall,
    PrimitiveType,
    PulseVoltageSourceParams,
)


def to_proto(
    top: Elaboratables,
    domain: Optional[str] = None,
    **kwargs,
) -> vckt.Package:
    """Convert Elaborate-able Module or Generator `top` and its dependencies to a Proto-format `Package`."""
    # Elaborate all the top-level Modules
    tops = elaborate(top)
    if not isinstance(tops, list):
        tops = [tops]
    exporter = ProtoExporter(tops=tops, domain=domain)
    return exporter.export()


@datatype(config=AllowArbConfig)
class ModuleMapping:
    hmod: Module  # hdl21.Module
    pmod: vckt.Module  # VLSIR Module


class ProtoExporter:
    """
    Hierarchical Protobuf Exporter
    Walks a Module hierarchy from Module `self.top`, defining each required Module along the way.
    Modules are defined in `self.pkg` in dependency order.
    Upon round-tripping, all dependent child-modules will be encountered before their parent instantiators.
    """

    def __init__(self, tops: List[Module], domain: Optional[str] = None):
        self.tops = tops

        # Module mappings, keyed by (a) hdl21.Module (ID), and (b) vlsir.Module (name)
        self.modules_by_id: Dict[int, ModuleMapping] = dict()
        self.modules_by_name: Dict[str, ModuleMapping] = dict()

        # ExternalModule-id to Proto-ExternalModule dict
        self.ext_modules: Dict[int, vckt.ExternalModule] = dict()

        # Default `domain` AKA package-name is the empty string
        self.pkg = vckt.Package(domain=domain or "")

    def export(self) -> vckt.Package:
        """Export starting with every Module in `self.tops`,
        visiting every hierarchical node along the way.
        Returns our generated `Package` as a result."""
        if not isinstance(self.tops, list):
            raise TypeError
        for m in self.tops:
            if not isinstance(m, Module):
                raise TypeError
            self.export_module(m)
        return self.pkg

    def export_module_name(self, module: Module) -> vlsir.utils.QualifiedName:
        """Create and return a unique `QualifiedName` for Module `module`.
        Raises a `RuntimeError` if unique name is taken."""

        mname = module_qualname(module)

        if mname in self.modules_by_name:
            conflict = self.modules_by_name[mname].hmod
            msg = f"Cannot serialize Module {module} due to conflicting name with {conflict}. \n"
            msg += "(Was this a generator that didn't get decorated with `@hdl21.generator`?) "
            raise RuntimeError(msg)
        return mname

    def export_module(self, module: Module) -> vckt.Module:
        """Export a Module `module` and all its dependencies."""

        if id(module) in self.modules_by_id:  # Already done
            return self.modules_by_id[id(module)].pmod

        if module.bundles:  # Invalid, should have been elaborated out.
            msg = f"Invalid attribute for Proto export: Module {module.name} with Bundles {list(module.bundles.keys())}"
            raise RuntimeError(msg)

        # Create the Proto-Module
        pmod = vckt.Module()

        # Create its serialized name
        pmod.name = self.export_module_name(module)

        # Create its Signal-objects, which include the hdl21.Module's Ports
        for sig in list(module.signals.values()) + list(module.ports.values()):
            psig = vckt.Signal(name=sig.name, width=sig.width)
            pmod.signals.append(psig)

        # Create its Port-objects
        for port in module.ports.values():
            pmod.ports.append(export_port(port))

        # Create each Proto-Instance
        for inst in module.instances.values():
            if not inst.of:
                msg = f"Invalid Instance {inst.name} of unresolved Module in Module {module.name}"
                raise RuntimeError(msg)
            pinst = self.export_instance(inst)
            pmod.instances.append(pinst)

        # Create the Module's `literal`s
        for literal in module.literals:
            pmod.literals.append(export_literal(literal))

        # Store references to the result, and return it
        mapping = ModuleMapping(module, pmod)
        self.modules_by_id[id(module)] = mapping
        self.modules_by_name[pmod.name] = mapping
        self.pkg.modules.append(pmod)
        return pmod

    def export_external_module(self, emod: ExternalModule) -> vckt.ExternalModule:
        """Export an `ExternalModule`"""
        if id(emod) in self.ext_modules:  # Already done
            return self.ext_modules[id(emod)]

        # ...
        pmod = export_external_module(emod)

        # Store references to the result, and return it
        self.ext_modules[id(emod)] = pmod
        self.pkg.ext_modules.append(pmod)
        return pmod

    def export_instance(self, inst: Instance) -> vckt.Instance:
        """Convert an hdl21.Instance into a Proto-Instance
        Depth-first retrieves a Module definition first,
        using its generated `name` field as the Instance's `module` pointer."""

        # Create the Proto-Instance
        pinst = vckt.Instance(name=inst.name)

        # First depth-first seek out our definition,
        # Retrieving the data we need to make a `Reference` to it
        if isinstance(inst.of, Module):
            pmod = self.export_module(inst.of)
            # Give it a Reference to its Module
            pinst.module.local = pmod.name
        elif isinstance(inst.of, (PrimitiveCall, ExternalModuleCall)):
            call = inst.of

            if isinstance(inst.of, PrimitiveCall):
                # Create a reference to one of the `primitive` namespaces
                if call.prim.primtype == PrimitiveType.PHYSICAL:
                    # FIXME: #54 also expose the `hdl21.primitives` as a VLSIR package
                    pinst.module.external.domain = "hdl21.primitives"
                    pinst.module.external.name = call.prim.name
                    params = dictify_params(call.params)
                elif call.prim.primtype == PrimitiveType.IDEAL:
                    # Ideal elements convert to `vlsir.primitives`
                    pinst.module.external.domain = "vlsir.primitives"
                    prim_map = {
                        "DcVoltageSource": "vdc",
                        "PulseVoltageSource": "vpulse",
                        "SineVoltageSource": "vsin",
                        "CurrentSource": "isource",
                        "IdealResistor": "resistor",
                        "IdealCapacitor": "capacitor",
                        "IdealInductor": "inductor",
                        "VoltageControlledVoltageSource": "vcvs",
                        "CurrentControlledVoltageSource": "ccvs",
                        "VoltageControlledCurrentSource": "vccs",
                        "CurrentControlledCurrentSource": "cccs",
                    }
                    if call.prim.name not in prim_map:
                        msg = f"Invalid Primitive {call.prim.name} in PrimitiveCall {inst.name}"
                        raise RuntimeError(msg)
                    pinst.module.external.name = prim_map[call.prim.name]
                    params = export_primitive_params(call.params)
                else:
                    raise ValueError(f"Invalid PrimitiveType {call.prim.primtype}")

            elif isinstance(inst.of, ExternalModuleCall):
                self.export_external_module(call.module)
                pinst.module.external.domain = call.module.domain or ""
                pinst.module.external.name = call.module.name
                params = dictify_params(call.params)

            else:
                msg = (
                    f"Un-exportable Instance {inst} resolves to invalid type {inst.of}"
                )
                raise TypeError(msg)

            # Export the instance parameters
            for key, val in params.items():
                if val is None:
                    continue  # None-valued parameters go un-set
                # Otherwise export and copy it into place
                vparam = vlsir.Param(name=key, value=export_param_value(val))
                pinst.parameters.append(vparam)

        else:
            raise TypeError(f"Un-exportable Instance {inst} of {inst.of}")

        # Create its connections mapping
        for pname, conn in inst.conns.items():
            # Assign each item into the connections dict.
            # The proto interface requires copying it along the way
            pconn = vckt.Connection(
                portname=pname, target=export_connection_target(conn)
            )
            pinst.connections.append(pconn)

        return pinst


def export_port(port: Port) -> vckt.Port:
    """Export a `Port`"""
    pport = vckt.Port()
    pport.direction = export_port_dir(port)
    pport.signal = port.name
    return pport


def export_port_dir(port: Port) -> vckt.Port.Direction:
    """Convert between Port-Direction Enumerations"""
    if port.direction == PortDir.INPUT:
        return vckt.Port.Direction.INPUT
    if port.direction == PortDir.OUTPUT:
        return vckt.Port.Direction.OUTPUT
    if port.direction == PortDir.INOUT:
        return vckt.Port.Direction.INOUT
    if port.direction == PortDir.NONE:
        return vckt.Port.Direction.NONE
    raise ValueError(f"Invalid PortDir {port.direction}")


def export_connection_target(
    sig: Union[Signal, Slice, Concat]
) -> vckt.ConnectionTarget:
    """Export a proto `ConnectionTarget`"""

    pconn = vckt.ConnectionTarget()  # Create a proto-Connection
    if isinstance(sig, Signal):
        pconn.sig = sig.name
    elif isinstance(sig, Slice):
        pslice = export_slice(sig)
        pconn.slice.CopyFrom(pslice)
    elif isinstance(sig, Concat):
        pconc = export_concat(sig)
        pconn.concat.CopyFrom(pconc)
    else:
        msg = f"Invalid argument to `export_connection_target`: {sig}"
        raise TypeError(msg)
    return pconn


def export_slice(slize: Slice) -> vckt.Slice:
    """Export a signal-`Slice`.
    Fails if the parent is not a concrete `Signal`, i.e. it is a `Concat` or another `Slice`.
    Fails for non-unit step-sizes, which should be converted to `Concat`s upstream."""
    if not isinstance(slize.parent, Signal):
        msg = f"Export error: {slize} has a parent {slize.parent} which is not a concrete Signal"
        raise RuntimeError(msg)
    if slize.step != 1:
        msg = f"Export error: {slize} has non-unit step"
        raise RuntimeError(msg)

    # Move to HDL-style indexing, with inclusive `top` index.
    return vckt.Slice(signal=slize.parent.name, top=slize.top - 1, bot=slize.bot)


def export_concat(concat: Concat) -> vckt.Concat:
    """Export (potentially recursive) Signal Concatenations"""
    pconc = vckt.Concat()
    for part in concat.parts:
        pconc.parts.append(export_connection_target(part))
    return pconc


def export_primitive_params(params: Any) -> Dict[str, Optional[Prefixed]]:
    """Convert the parameters of an `IDEAL` primitive into their VLSIR form.
    Returns the result as a dictionary of {name: value}s."""

    if isinstance(params, PulseVoltageSourceParams):
        return dict(
            v1=params.v1,
            v2=params.v2,
            td=params.delay,
            tr=params.rise,
            tf=params.fall,
            tpw=params.width,
            tper=params.period,
        )

    # For everything else, pass along the values name-by-name as a dictionary.
    return dictify_params(params)


def dictify_params(params: Any) -> Dict[str, Optional[Prefixed]]:
    """Turn a `paramclass` of parameters into a dictionary of exportable values.
    This is essentially a one-layer replacement for the standard library's
    `dataclasses.as_dict`, which we replace for sake of our "near-scalar" types
    such as `Prefixed` numbers."""

    if isinstance(params, dict):
        return params

    if not isparamclass(params):
        msg = f"Invalid Parameter Call-Argument {params} - must be either `dict` or `hdl21.paramclass`"
        raise TypeError(msg)

    # And turn its fields into a dictionary
    return {field.name: getattr(params, field.name) for field in fields(params)}


# Alias for the union of (python) types which can be converted to Vlsir's `ParamValue`.
# These can generally be summarized as "numbers, strings, and things easily convertible into them".
# Hdl21 parameters can of course be a much larger set of types.
# Anything outside this list produces a `TypeError` when exported.
ToVlsirParam = Union[
    type(None), int, float, str, Decimal, Enum, Prefixed, Scalar, Literal
]


def export_param_value(val: ToVlsirParam) -> Optional[vlsir.ParamValue]:
    """Export a `ParamValue`."""

    if isinstance(val, type(None)):
        return None  # `None` serves as the null identifier for "no value".

    if isinstance(val, str):
        # FIXME: the debate between `string` and `literal`. For now, we use LITERAL.
        return vlsir.ParamValue(literal=val)
    # Enum-valued parameters must also be strings, or fail
    if isinstance(val, Enum):
        if not isinstance(val.value, str):
            raise TypeError(f"Enum-valued parameters must be strings, not {val.value}")
        return vlsir.ParamValue(literal=val.value)

    # Internal numeric (and number-like) types
    if isinstance(val, Literal):  # String/ expression literals
        return vlsir.ParamValue(literal=val.text)
    if isinstance(val, Prefixed):
        return vlsir.ParamValue(prefixed=export_prefixed(val))

    # Standard-Lib Numbers
    if isinstance(val, Decimal):
        return vlsir.ParamValue(literal=str(val))
    if isinstance(val, int):
        return vlsir.ParamValue(int64_value=val)
    if isinstance(val, float):
        return vlsir.ParamValue(double_value=val)

    msg = f"Unsupported parameter for proto-export: `{val}`"
    raise TypeError(msg)


def export_prefix(pre: Prefix) -> vlsir.SIPrefix:
    """Export an enumerated `Prefix`"""
    map = {
        -24: vlsir.SIPrefix.YOCTO,
        -21: vlsir.SIPrefix.ZEPTO,
        -18: vlsir.SIPrefix.ATTO,
        -15: vlsir.SIPrefix.FEMTO,
        -12: vlsir.SIPrefix.PICO,
        -9: vlsir.SIPrefix.NANO,
        -6: vlsir.SIPrefix.MICRO,
        -3: vlsir.SIPrefix.MILLI,
        -2: vlsir.SIPrefix.CENTI,
        -1: vlsir.SIPrefix.DECI,
        0: vlsir.SIPrefix.UNIT,
        1: vlsir.SIPrefix.DECA,
        2: vlsir.SIPrefix.HECTO,
        3: vlsir.SIPrefix.KILO,
        6: vlsir.SIPrefix.MEGA,
        9: vlsir.SIPrefix.GIGA,
        12: vlsir.SIPrefix.TERA,
        15: vlsir.SIPrefix.PETA,
        18: vlsir.SIPrefix.EXA,
        21: vlsir.SIPrefix.ZETTA,
        24: vlsir.SIPrefix.YOTTA,
    }
    if pre.value not in map:
        raise ValueError(f"Invalid Prefix {pre}")
    return map[pre.value]


def export_prefixed(pref: Prefixed) -> vlsir.Prefixed:
    """Export a `Prefixed` number"""

    # Export the metric prefix
    prefix = export_prefix(pref.prefix)

    # And export the numeric part. Use Vlsir's `integer` variant for Decimal values which equal integers, and strings otherwise.
    if pref.number == int(pref.number):
        return vlsir.Prefixed(int64_value=int(pref.number), prefix=prefix)
    return vlsir.Prefixed(string_value=str(pref.number), prefix=prefix)


# FIXME: #54 also expose the `hdl21.primitives` as a VLSIR package
# @classmethod
# def export_hdl21_primitive(cls, prim: Primitive) -> vckt.ExternalModule:
#     """ Export a `Primitive` as a `vckt.ExternalModule`.
#     Not typically done as part of serialization,
#     but solely as an aid to other conversion utilities. """
#
#     # Create the Proto-ExternalModule
#     if prim.primtype == PrimitiveType.PHYSICAL:
#         domain = "hdl21.primitives"
#     elif prim.primtype == PrimitiveType.IDEAL:
#         domain = "hdl21.ideal"
#     else:
#         raise ValueError
#     qname = vlsir.utils.QualifiedName(name=prim.name, domain=domain)
#     pmod = vckt.ExternalModule(name=qname)
#
#     # Create its Port-objects
#     for port in prim.port_list:
#         pmod.ports.append(cls.export_port(port))
#
#     # And return it
#     return pmod


def export_literal(literal: Literal) -> str:
    """Export a `Literal`, as its text value"""
    return literal.text


def export_external_module(emod: ExternalModule) -> vckt.ExternalModule:
    """Export an `ExternalModule`"""

    # Create the Proto-ExternalModule
    qname = vlsir.utils.QualifiedName(name=emod.name, domain=emod.domain)
    pmod = vckt.ExternalModule(name=qname, spicetype=emod.spicetype.to_schema())

    # Create its Port-objects, which also require Vlsir Signal objects
    for port in emod.port_list:
        psig = vckt.Signal(name=port.name, width=port.width)
        pmod.signals.append(psig)
        pmod.ports.append(export_port(port))

    return pmod



""" 
# Elaboration Passes
"""

from enum import Enum
from typing import List

# Import all the defined passes
from .passes import (
    InstBundleElabPass,
    Orphanage,
    ConnTypes,
    BundleFlattener,
    ResolvePortRefs,
    ArrayFlattener,
    SliceResolver,
    MarkModules,
)


class ElabPass(Enum):
    """
    # Enumerated ElabPass Passes

    Each has a `value` attribute which is an `ElabPass` class,
    and a `name` attribute which is a (Python-enum-style) capitalized name.

    Typical usage involves arranging several `ElabPass` in a list
    to be performed in-order.
    The `default` class-method produces the default such list.
    """

    ORPHANAGE = Orphanage
    INSTANCE_BUNDLES = InstBundleElabPass
    RESOLVE_PORT_REFS = ResolvePortRefs
    CONN_TYPES = ConnTypes
    FLATTEN_BUNDLES = BundleFlattener
    FLATTEN_ARRAYS = ArrayFlattener
    RESOLVE_SLICES = SliceResolver
    MARK_MODULES = MarkModules

    @classmethod
    def default(cls) -> List["ElabPass"]:
        """Return the default ordered ElabPass Passes."""
        # Returns each in definition order, then a final few tests.
        return list(ElabPass)[:-1] + [
            ElabPass.CONN_TYPES,
            ElabPass.ORPHANAGE,
            ElabPass.MARK_MODULES,
        ]

    @property
    def elaborate(self):
        return self.value.elaborate
